#!/usr/bin/env python3
"""
Verifies that all assets are covered by a threat
"""
from data import find
import logging
import verifiers.reference as reference
from language.translate import Translate
from verifiers.verifier_error import ErrorType
from verifiers.verifier_error import VerifierIssue
from utils import match, transform
from utils import tags

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

   
def location_storage_expression_callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value):

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    if tag_comparison == "storage-expression":
        
        storage_location_value = callback_config.get("storage_location_value")
        if match.equals(storage_location_value, compare_to_value):

            #grouped_text = callback_config.get("grouped-text", {}).get("storage-expression")
            component_transform = callback_config.get("component-transform", transform.identity)

            if match.starts_ends(compare_value, Translate.localise(callback_config.get("output-texts", {}), "start-assets-grouped-by-storage", ignore_format=True), component_transform(compare_to_value)):
                return True
            if match.equals(compare_value, Translate.localise(callback_config.get("output-texts", {}), "all-assets", ignore_format=True)):
                return True
            #if match.endswith(compare_value, compare_to_value):
            #    return True

    return False

def component_storage_expression_callback(callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value):

    tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

    if tag_comparison == "storage-expression":
        #grouped_text = callback_config.get("grouped-text", {}).get("storage-expression")
        component_transform = callback_config.get("component-transform", transform.identity)

        if match.starts_ends(compare_value, Translate.localise(callback_config.get("output-texts", {}), "start-assets-grouped-by-storage", ignore_format=True), component_transform(compare_to_value)):
            return True
        if match.equals(compare_value, Translate.localise(callback_config.get("output-texts", {}), "all-assets", ignore_format=True)):
            return True

    return False

def template_reference_callback(callback_config, tag_comparison, compare_value, compare_to_key, compare_to_value):

    return False

def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = assets_verify(common_config, verifier_config, model, template_model)

    # Not currently used.  Likely to be deprecated.  Too complicated.
    #verify_return_list.extend(cov_verify(common_config, verifier_config, model, template_model))

    return verify_return_list

def assets_verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:
    verify_return_list = []

    # Get all the data required to do the validation
    tagged_asset_data_list = common_config["asset-tags"]["asset-data-tag"]
    all_assets = []
    for tag in tagged_asset_data_list:
        tag_asset_data_key, tag_asset_data_value = find.key_with_tag(model, tag)
        if not isinstance(tag_asset_data_value, list):
            logger.error(f"Expecting the data tagged with '{tag}' to be a list, but it was a '{type(tag_asset_data_value)}'. Ignoring.")
            continue
    
        all_assets.append((tag, tag_asset_data_key, tag_asset_data_value))

    if len(all_assets) == 0:
        logger.error(f"Could not find any asset data tagged with '{common_config['asset-tags']['asset-data-tag']}'")

    threats_and_controls_key, threats_and_controls_data = find.key_with_tag(model, common_config["threat-tags"]["threats-data-tag"])
    if threats_and_controls_data is None:
        logger.error(f"Could not find threat data tagged with '{common_config['threat-tags']['threats-data-tag']}'")

    # Get a list of in-scope and out-of-scope components
    in_scope_components = []
    out_of_scope_components = []
    components_key, components_data = find.key_with_tag(model, common_config["component-tags"]["component-data-tag"])
    for component_row in components_data:
        component_row_id_key, component_row_id_data = find.key_with_tag(component_row, "row-identifier")
        in_scope_key, in_scope_data = find.key_with_tag(component_row, common_config["component-tags"]["component-in-scope-tag"])
        if match.equals(in_scope_data, common_config["component-tags"]["component-in-scope-value"]):
            in_scope_components.append(component_row_id_data)
        else:
            out_of_scope_components.append(component_row_id_data)

    component_transform = transform.strip(common_config["strip-context"]["start-char"], common_config["strip-context"]["end-char"])

    #callback_config = {"grouped-text":common_config["grouped-text"], "component-transform":component_transform}
    callback_config = {"output-texts":common_config["output-texts"], "component-transform":component_transform}

    exclude_callback = lambda callback_config, tag_tuple, compare_value, compare_to_key, compare_to_value: match.contains(compare_value, compare_to_value)

    # Loop through the asset data tables
    for asset_data_tag, asset_data_key, asset_data_value in all_assets:

        # Loop through the actual rows of assets in a data table
        for asset in asset_data_value:

            # Get the row identifier for the asset (to store covering threats against and for error reporting)
            row_id_key, row_id_data = find.key_with_tag(asset, "row-identifier")
            if row_id_key is None:
                logger.error(f"Could not find a 'row-identifier' tag on any field for asset '{asset}'")

            # Get the storage-locations for the asset
            for storage_location_key, storage_location_value in find.keys_with_tag(asset, common_config["asset-tags"]['asset-location-tag']):

                if match.equals(storage_location_value, out_of_scope_components, component_transform):
                    logger.debug(f"Ignoring threat coverage for asset '{row_id_key.name}' in storage location '{storage_location_value}' as '{storage_location_value}' is out of scope")
                    continue

                # Ignore any excluded values
                exclude = False
                for exclude_tag in tags.get_prefixed_tag(common_config["exclude"]["tag-prefix"] + "/", storage_location_key):
                    exclude_tag_tuple = tags.get_quad_tag_parts(exclude_tag)
                    tag_prefix, tag_type, tag_value, tag_comparison = exclude_tag_tuple
                    
                    if exclude := tags.check_tag_comparison(exclude_tag_tuple, storage_location_value, storage_location_key, tag_value, exclude_callback, None, True):
                        storage_location_key.addProperty("excluded", True)
                        logger.debug(f"Ignoring threat coverage for asset '{row_id_key.name}' in storage location '{storage_location_value}' as tag '{exclude_tag}' excludes it")
                        break
                if exclude:
                    continue

                logger.debug(f"Checking threat coverage for asset '{row_id_key.name}' in storage location '{storage_location_value}'")

                # Add an empty list of covering threats.  The list will be appended to as it is found.
                covering_threats = []
                row_id_key.addProperty(storage_location_key, covering_threats)

                # An asset is covered by a threat when:
                # Scenario A: threat component includes = in-scope asset storage location AND asset = name
                # Scenario B: threat component includes = in-scope asset storage location AND asset = type
                # Scenario C: threat component includes = in-scope asset storage location AND asset = all assets in component
                # Scenario D: threat component is anything AND asset = all assets in known storage type (e.g. env var)
                
                # Try to find a covering threat for the asset in the storage location by searching the threats and controls data
                for entry in threats_and_controls_data:

                    # An entry in the threats data section can contain multiple assets and components
                    
                    # Note, we are just searching the threats row here, so the assets and components returned will be in the same table row
                    threat_asset_entries = find.keys_with_tag(entry, common_config["threat-tags"]["threat-asset-tag"])
                    threat_component_entries = find.keys_with_tag(entry, common_config["threat-tags"]["threat-component-tag"])

                    if len(threat_asset_entries) == 0:
                        # This validation issue will be raised elsewhere
                        continue

                    # Does the storage location for the asset match one of the components?
                    matching_component = match.get_equals(storage_location_value, [component_name for (_, component_name) in threat_component_entries], component_transform)
                    if match.equals(matching_component, in_scope_components, component_transform):
                        matching_in_scope_component = matching_component
                    else:
                        matching_in_scope_component = None

                    for threat_asset_entry_key, threat_asset_entry_value in threat_asset_entries:

                        if matching_in_scope_component is not None:
                            # Scenario A, B, C
                            # Then the asset can match on name or type, but the storage location must match the component
                            
                            # This checks if any 'ref' tag in 'threat_asset_entry_key' can be found in row 'asset' for value 'threat_asset_entry_value'
                            if reference.check_reference_row(asset, "ref", threat_asset_entry_key, threat_asset_entry_value, component_storage_expression_callback, callback_config, only_callback=False) is not None and match.equals(matching_in_scope_component, storage_location_value, component_transform):
                                if entry not in covering_threats:
                                    covering_threats.append(entry)
                        else:
                            # Scenario D
                            # The asset can only match on grouped storage type e.g. All assets stored in env vars
                            callback_config["storage_location_value"] = storage_location_value

                            if (matched_tag := reference.check_reference_row(asset, "ref", threat_asset_entry_key, threat_asset_entry_value, location_storage_expression_callback, callback_config, only_callback=True)) is not None and match.endswith(matched_tag, "storage-expression"):
                                if entry not in covering_threats:
                                    covering_threats.append(entry)

                if len(covering_threats) == 0:
                    issue_dict = {}
                    issue_dict["issue_key"] = row_id_key
                    issue_dict["issue_value"] = row_id_data
                    issue_dict["issue_table"] = threats_and_controls_key.getProperty("section")
                    issue_dict["issue_table_row"] = None    # Disable showing this in the error message
                    #issue_dict["fixdata"] = Translate.localise(common_config["grouped-text"]["storage-expression"], "start-assets-grouped-by-storage")
                    #issue_dict["fixdata"] = common_config["grouped-text"]["storage-expression"]
                    # Need to get some helpful texts relating to the threats and control data
                    # Although we found the issue by looping over asset keys, the problem is in the threats and controls data
                    issue_dict["storage_location"] = storage_location_value
                    issue_dict["threats_and_controls_table"] = threats_and_controls_key.getProperty("section")
                    
                    issue_dict["asset_key"] = None
                    # We need to get information about the key tagged "threat-asset-tag" but have to handle the scenario where no entries have yet been put in the column 
                    if len(threats_and_controls_data) > 0:
                        threat_asset_entries = find.keys_with_tag(threats_and_controls_data, common_config["threat-tags"]["threat-asset-tag"])
                        if len (threat_asset_entries) > 0:
                            for keyentry, valueentry in threat_asset_entries:
                                if not keyentry is None:
                                    issue_dict["asset_key"] = keyentry
                                    break
                    
                    if issue_dict["asset_key"] is None:
                        # We could find a key tagged with "threat-asset-tag"
                        fix_text_key = "no-covering-threat-fix-no-threat-or-asset"
                    else:
                        fix_text_key = "no-covering-threat-fix"

                    verify_return_list.append(VerifierIssue(
                        error_text_key="no-covering-threat", 
                        fix_text_key=fix_text_key, 
                        fix_data_key="start-assets-grouped-by-storage",
                        issue_dict=issue_dict, 
                        errortype=ErrorType.NOT_SET))
            
                else:
                    # Store covering threats
                    storage_location_key.addProperty("covering-threats", covering_threats)

    return verify_return_list

in_scope_components = []
out_of_scope_components = []

def _filter(common_config, model, comparison, covered_key_values):

    filtered_keyvalues = []

    if match.equals(comparison, "equals"):
        return covered_key_values

    if match.equals(comparison, "component_in_scope"):
        if len(in_scope_components) == 0 and len(out_of_scope_components) == 0:
            components_key, components_data = find.key_with_tag(model, common_config["component-tags"]["component-data-tag"])
            for component_row in components_data:
                component_row_id_key, component_row_id_data = find.key_with_tag(component_row, "row-identifier")
                in_scope_key, in_scope_data = find.key_with_tag(component_row, common_config["component-tags"]["component-in-scope-tag"])
                if match.equals(in_scope_data, common_config["component-tags"]["component-in-scope-value"]):
                    in_scope_components.append(component_row_id_data)
                else:
                    out_of_scope_components.append(component_row_id_data)
        
        for key_entry, value_entry in covered_key_values:
            if match.equals(value_entry, in_scope_components):
                filtered_keyvalues.append((key_entry, value_entry))

    return filtered_keyvalues

def cov_verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = []

    # Get all the "cov" references
    # Find every key with a tag that starts with the configured prefix.
    all_cov_tagged_keys = find.keys_with_tag_matching_regex(model, "^cov/.*$")

    all_cov_tagged_keys = [(key, value, tag) for key, value in all_cov_tagged_keys for tag in tags.get_prefixed_tag("cov", key)]
               
    # Group by 'covered' context and 'covering' context
    covered_tagged_keys = [(key, value, tag) for key, value, tag in all_cov_tagged_keys if (tags.get_quad_tag_parts(tag)[1]).startswith("covered") ]
    covering_tagged_keys = [(key, value, tag) for key, value, tag in all_cov_tagged_keys if (tags.get_quad_tag_parts(tag)[1]).startswith("covering") ]

    # For each 'covered' context identifier
    for covered_section_key, covered_section_value, tag in covered_tagged_keys:

        if (section := covered_section_key.getProperty("section")) is None:
            logger.warning(f"Coverage tag '{tag}' was not applied to section.  Ignoring")
            continue

        prefix, context, covered_key_tag, comparison = tags.get_quad_tag_parts(tag)
        if len(context.split("=")) != 2:
            logger.warning(f"Found coverage tag '{context}' with context that didn't include an ID")
            continue
        covered_context_id = context.split("=")[1]

        # Get the covered tagged keys
        covered_key_values = find.keys_with_tag(covered_section_value, covered_key_tag)

        # Filter the covered tagged keys
        covered_key_values = _filter(common_config, model, comparison, covered_key_values)

        # For each 'covering' context with same identifier
        for covering_section_key, covering_section_value, tag in covering_tagged_keys:

            if (section := covering_section_key.getProperty("section")) is None:
                logger.warning(f"Coverage tag '{tag}' was not applied to section.  Ignoring")
                continue

            prefix, context, covering_key_tag, comparison = tags.get_quad_tag_parts(tag)
            if len(context.split("=")) != 2:
                logger.warning(f"Found coverage tag '{context}' with context that didn't include an ID")
                continue
            covering_context_id = context.split("=")[1]

            if not match.equals(covered_context_id, covering_context_id):
                continue

            # Get the covering tagged keys
            covering_key_values = find.keys_with_tag(covering_section_value, covering_key_tag)

            # Filter the covering tagged keys
            covering_key_values_filtered = _filter(common_config, model, comparison, covering_key_values)

            # Check if there is a covering tagged key value that matches each covered tagged key value
            covering_values = [value_entry for key_entry, value_entry in covering_key_values_filtered]

            for key_entry, value_entry in covered_key_values:

                if not match.equals(value_entry, covering_values):
                    # Coverage finding
                    issue_dict = {}
                    issue_dict["issue_key"] = key_entry
                    issue_dict["issue_value"] = value_entry
                    issue_dict["issue_table"] = covering_section_key.getProperty("section")
                    issue_dict["issue_table_row"] = None    # Disable showing this in the error message
                    if len(covering_key_values) > 0:
                        key_eg, _ = covering_key_values[0]
                        issue_dict["covering_col"] = key_eg.getProperty("colname")
                    else:
                        # Best we can do when there is no example of a covering key value
                        issue_dict["covering_col"] = covering_key_tag

                    issue_dict["covering_table"] = covering_section_key.getProperty("section")
                    
                    verify_return_list.append(VerifierIssue(
                        error_text_key=covered_context_id, 
                        fix_text_key=covered_context_id + "-fix", 
                        issue_dict=issue_dict, 
                        errortype=ErrorType.NOT_SET))

    return verify_return_list


#!/usr/bin/env python3
"""
Validates whether a value matches a regular expression
"""

import logging
from pathlib import Path
import re
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def validate(config:dict, key:Key, value:str, references:dict) -> bool:

    if not (pattern := config.get("pattern", None)):
        logger.error("Validator config did not have a 'pattern' key")
        return False

    try:
        if re.search(pattern, value):
            return True
    except re.error as err:
        logger.error(f"The pattern '{pattern}' passed to the regex validator caused an exception '{err}'")
        return False

    return False


#!/usr/bin/env python3
"""
Verifies that all kets tagged as mandatory have a non-empty value, if some dependent key also has a value
"""

from data import find
import logging
from verifiers.verifier_error import VerifierIssue
from utils import match
from utils import keymaster

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:

    verify_return_list = []

    # The config could contain many tagged values to look for
    if (tags := verifier_config.get("tags")) is None:
        return verify_return_list

    for tag_entry in verifier_config.get("tags"):
        
        # Need to get all keys for a given tag
        tagged_data = find.keys_with_tag(model, tag_entry["tag"])
        #print(f"tagged_data = {tagged_data}")

        for tagged_data_entry_key, tagged_data_entry_value in tagged_data:
            # Get the rowID for the tagged key
            rowIDkey = keymaster.get_row_identifier_for_key(tagged_data_entry_key)
            row_data = rowIDkey.getProperty("row")
            depends_on_key, depends_on_key_value = find.key_with_tag(row_data, tag_entry["depends-on"])
            # 1st clause: does the depend_on value match, and is the tagged_key value empty -> then finding
            # 2nd clause: is there no depend_on value specified, and if the tagged_key value is not empty (i.e. it has a value) -> then finding
            if (("depends-on-value" in tag_entry and match.equals(depends_on_key_value, tag_entry["depends-on-value"])) and match.is_empty(tagged_data_entry_value)) or \
                (("depends-on-value" not in tag_entry and not match.is_empty(depends_on_key_value)) and match.is_empty(tagged_data_entry_value)) :
                issue_dict = {}
                issue_dict["issue_key"] = tagged_data_entry_key
                issue_dict["issue_value"] = tagged_data_entry_value
                issue_dict["depends_on_key"] = depends_on_key
                issue_dict["depends_on_value"] = depends_on_key_value
                if "depends-on-value" in tag_entry:
                    verify_return_list.append(VerifierIssue(
                        error_text_key="missing-conditional-mandatory", 
                        fix_text_key=None, 
                        issue_dict=issue_dict))
                else:
                    verify_return_list.append(VerifierIssue(
                        error_text_key="missing-conditional-mandatory-any-value", 
                        fix_text_key=None, 
                        issue_dict=issue_dict))


        # # Since we need to look at adjacent fields for the condition, get a list of all sections containing tagged data.  A section is like a table
        # tagged_data_section_list = []
        # for tagged_data_entry_key, tagged_data_entry_value in tagged_data:

        #     # TODO: adding sections is not mandatory, so using them isn't reliable
        #     # Get section that contains that tag
        #     sectionKey = keymaster.get_section_for_key(tagged_data_entry_key)
        #     if sectionKey is not None and sectionKey.getProperty("section") is not None and sectionKey not in tagged_data_section_list:
        #         tagged_data_section_list.append(sectionKey)
        # #print(f"tagged_data_section_list = {tagged_data_section_list}")

        # # For each list entry in the section, find the 2 fields that depend on each other (so we are assuming each section contains a list or dict)
        # for data_section in tagged_data_section_list:
        #     if not (data_section_value := data_section.getProperty("value")):
        #         logger.error(f"Discovered section {data_section} did not have a property 'value'")
        #         continue

        #     if isinstance(data_section_value, dict):
        #         data_section_value = [data_section_value]
        #     #print(f"data_section_value = {data_section_value}")
        #     # The section must be a list and the list entries must be dicts, one key of which must be the tagged data
        #     for dict_entry in data_section_value:
        #         if tag_entry["tag"] in dict_entry and tag_entry["depends-on"] in dict_entry:
        #             # Check if the 'depends-on' field has the necessary value and check the target value exists
        #             if match.equals(dict_entry[tag_entry["depends-on"]], tag_entry["depends-on-value"]) and \
        #                 match.is_empty(dict_entry[tag_entry["tag"]]):
        #                 # Need to get the tagged key to give to the VerifierError
        #                 dict_key_list = list(dict_entry.keys())
        #                 tagged_key = dict_key_list.pop(dict_key_list.index(tag_entry["tag"]))

        #                 verify_return_list.append(VerifierError(verifier_config, 
        #                                             verifier_name,
        #                                             errorTexts["missing-conditional-mandatory"].format(tag_entry["depends-on"], tag_entry["depends-on-value"], tag_entry["tag"]), 
        #                                             tagged_key))

    return verify_return_list


#!/usr/bin/env python3
"""
Class ManageOutput
"""

import logging
import utils.load_yaml
from language.translate import Translate
import jsonpickle

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# from jinja2 import Environment, FileSystemLoader, select_autoescape
# env = Environment(
#     loader = FileSystemLoader(searchpath="./"),
#     autoescape=select_autoescape()
# )

class ManageOutput:

    def __init__(self, config):

        template_text_file = config.get("template-text-file")

        self.templated_texts = utils.load_yaml.yaml_file_to_dict(template_text_file).get("output-texts")

    def _getOutput(self, result, description, details = None):

        output = {}
        output["result"] = result
        output["description"] = description
        if details is not None:
            output["details"] = details

        return output

    def setInformation(self, text_key:str, template_values:dict, details = None) -> dict:

        info_text = Translate.localise(self.templated_texts, text_key, template_values)

        return self._getOutput("Information", info_text, details)

    def setSuccess(self, text_key:str, template_values:dict) -> dict:

        success_text = Translate.localise(self.templated_texts, text_key, template_values)

        return self._getOutput("Success", success_text, template_values["tm_version"])

    def setError(self, text_key:str, template_values:dict) -> dict:

        error_text = Translate.localise(self.templated_texts, text_key, template_values)

        return self._getOutput("Error", error_text)

    def tojson(self, output:dict):
        return jsonpickle.encode(output, unpicklable=False)


#!/usr/bin/env python3
"""
Class FormatOutput
"""

import logging
from enum import Enum
from pathlib import Path
from utils.config import ConfigBase
from utils import load_yaml
from utils.config import ConfigBase
from utils.request import Request
from language.translate import Translate
import jsonpickle

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

OUTPUT_TEXTS_YAML = "output_texts.yaml"
OUTPUT_TEXTS_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(OUTPUT_TEXTS_YAML))

class OutputType(Enum):
    ERROR = 0
    INFO = 1
    SUCCESS = 2
    NOT_SET = -1

    def __repr__(self):
        return self.name

class FormatOutput:

    #request_parameters:dict
    output_format:str = "json"
    #translator:Translate
    html_document:str = ""

    def __init__(self, output_config:dict):

        formatoutput_texts = load_yaml.yaml_file_to_dict(ConfigBase.getConfigPath(OUTPUT_TEXTS_YAML_PATH)).get("output-texts")
        self.information = Translate.localise(formatoutput_texts, "information")
        self.success = Translate.localise(formatoutput_texts, "success")
        self.error = Translate.localise(formatoutput_texts, "error")
        
        self.templated_texts:dict = load_yaml.yaml_file_to_dict(ConfigBase.getConfigPath(output_config.get("template-text-file"))).get("output-texts")
        # Need to merge these dicts at each localisation entry
        for dict_key in formatoutput_texts.keys():
            if dict_key in self.templated_texts:
                self.templated_texts[dict_key] = formatoutput_texts[dict_key] | self.templated_texts[dict_key]
            else:
                self.templated_texts[dict_key] = formatoutput_texts[dict_key]
        
        self.type = OutputType.NOT_SET
        self.description = None
        self.details = None

    def _get_state(self):

        output = {}
        if self.type == OutputType.ERROR:
            output["result"] = self.error
        elif self.type == OutputType.INFO:
            output["result"] = self.information
        elif self.type == OutputType.SUCCESS:
            output["result"] = self.success

        if self.description is not None:
            output["description"] = self.description
        if self.details is not None:
            output["details"] = self.details
        output["request"] = Request.get()

        return output

    def setInformation(self, text_key:str, template_values:dict, details = None):
        """ Returns a localised Information output message """

        if template_values is None:
            template_values = {}
            
        self.type = OutputType.INFO
        #self.description = Translate.localise(self.templated_texts, text_key, template_values | self.request_parameters)
        self.description = Translate.localise(self.templated_texts, text_key, template_values | {"request": Request.get()})
        self.details = details
        
        return

    def setSuccess(self, text_key:str, template_values:dict, details = None):
        """ Returns a localised Success output message """

        if template_values is None:
            template_values = {}

        self.type = OutputType.SUCCESS
        #self.description = Translate.localise(self.templated_texts, text_key, template_values | self.request_parameters)
        self.description = Translate.localise(self.templated_texts, text_key, template_values | {"request": Request.get()})
        self.details = details

        return

    def setError(self, text_key:str, template_values:dict, details = None):
        """ Returns a localised Error output message """

        if template_values is None:
            template_values = {}

        self.type = OutputType.ERROR
        #self.description = Translate.localise(self.templated_texts, text_key, template_values | self.request_parameters)
        self.description = Translate.localise(self.templated_texts, text_key, template_values | {"request": Request.get()})
        self.details = details

        return

    def getResult(self) -> OutputType:
        """ Returns an OutputType enum """
        return self.type

    def getDescription(self):
        """ Returns the description """
        return self.description

    def getDetails(self):
        """ Returns the details added to an Info/Success/Error message """
        return self.details

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    def tojson(self):
        return jsonpickle.encode(self, unpicklable=False)

    def toyaml(self):
        return load_yaml.class_to_yaml_str(self._get_state())

    def getContent(self, config_fn = None):

        if config_fn is not None:
            config_fn()

        if FormatOutput.output_format == "yaml":
            return ("text/yaml", self.toyaml())
        
        if FormatOutput.output_format == "html":
            return ("text/html", self.html_document)

        return ("application/json", self.tojson())
    
    @classmethod
    def setDocument(cls, document):

        cls.html_document = document


#!/usr/bin/env python3
"""
Loads validators dynamically
"""

from utils.error import ValidatorsError
import logging
import importlib
from pathlib import Path
from utils.config import ConfigBase
from language.translate import Translate
from utils.load_yaml import yaml_file_to_dict
from data.key import key as Key
from validators.validator_output import ValidatorOutput

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class Validator:
    """ Invokes a range of different configurable, extendable validation methods

    A class that uses 2 yaml files to configure and discover validation methods, that can then be invoked on values
    depending on the 'tags' associated with the value.
    """

    VALIDATORS_YAML = "validators_dispatch.yaml"
    VALIDATOR_CONFIG_YAML = "validators_config.yaml"
    VALIDATORS_TEXT_YAML = "validators_text.yaml"

    VALIDATORS_DISPATCH_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(VALIDATORS_YAML))
    VALIDATOR_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(VALIDATOR_CONFIG_YAML))
    VALIDATORS_TEXT_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(VALIDATORS_TEXT_YAML))

    def __init__(self, validators_config:dict = {}):
        #validators_dispatch_yaml_path = validators_config.get("validator-dispatch-yaml-path", self.VALIDATORS_DISPATCH_YAML_PATH)
        # We don't currently want the dispatch yaml to configurable - could be security implications as people who can update config shouldn't necessarily be people who can run code
        validators_dispatch_yaml_path = self.VALIDATORS_DISPATCH_YAML_PATH
        validators_config_yaml_path = ConfigBase.getConfigPath(validators_config.get("validator-config-yaml-path", self.VALIDATOR_CONFIG_YAML_PATH))
        validators_text_yaml_path = ConfigBase.getConfigPath(validators_config.get("validator-text-yaml-path", self.VALIDATORS_TEXT_YAML_PATH))
                
        self.text_dict = self._load_validator_texts(validators_text_yaml_path)
        self.validator_config_dict = self._load_validator_config(validators_config_yaml_path)
        self.dispatch, self.modules = self._load_validator_dispatch(validators_dispatch_yaml_path)
        
        ValidatorOutput.templated_output_texts = self.text_dict
        ValidatorOutput.validator_config = self.validator_config_dict
    

    def _load_validator_texts(self, validator_text_yaml_path) -> dict:

        yaml_config_dict = yaml_file_to_dict(validator_text_yaml_path) 

        if yaml_config_dict.get("validators-text") == None:
            raise ValidatorsError(f"Validators text file '{validator_text_yaml_path}' did not have a root key of 'validators-text'")

        # Convert to a dict for easy lookup
        validators_text_dict = {}
        for entry in yaml_config_dict["validators-text"]:
            if entry.get("tag") == None:
                raise ValidatorsError(f"The validators text file '{validator_text_yaml_path}' entry under the root 'validators-text' is missing a 'tag' key")
            validators_text_dict[entry["tag"]] = entry

        return validators_text_dict


    def _load_validator_config(self, validator_config_yaml_path) -> dict:

        #yaml_config_dict = yaml_templated_file_to_dict(validator_config_yaml_path, self.values_dict) 
        yaml_config_dict = Translate.localiseYamlFile(validator_config_yaml_path) 

        if yaml_config_dict.get("validator-config") == None:
            raise ValidatorsError(f"Validator config file '{validator_config_yaml_path}' did not have a root key of 'validator-config'")

        # Convert to a dict for easy lookup
        validator_config_dict = {}
        for entry in yaml_config_dict["validator-config"]:
            if entry.get("tag") == None:
                raise ValidatorsError(f"A validator config file '{validator_config_yaml_path}' entry under the root 'validator-config' is missing a 'tag' key")
            validator_config_dict[entry["tag"]] = entry

        return validator_config_dict


    def _load_validator_dispatch(self, validator_yaml_path) -> dict:

        #yaml_dict = yaml_to_dict(str(Path(os.getcwd()).joinpath(MAPS_DIR).joinpath(SCHEMES_YAML)))
        yaml_dict = yaml_file_to_dict(validator_yaml_path) 
        
        all_validators = yaml_dict["validators"]
        
        validators_dict = {}
        modules_dict = {}

        for validator_name in all_validators:
            validator_file = all_validators.get(validator_name, "")
            if not validator_file:
                logger.error("Could not find validator file '{}'".format(validator_name))
                continue
    
            module_name = validator_file
            if validator_file.endswith(".py"):
                module_name = Path(__file__).parent.stem + "." + Path(validator_file).stem
            logger.debug("Loading validator file '{}' as module '{}'".format(validator_file, module_name))
    
            # TODO catch ModuleNotFoundError exceptions
            imp = importlib.import_module(module_name)
            if hasattr(imp, "validate") and callable(imp.validate):
                validators_dict[validator_name] = imp.validate
                modules_dict[validator_name] = module_name
            else:
                logger.warning(f"Validator module/file '{module_name}' did not have a 'validate' method")
        
        return validators_dict, modules_dict



    def validate(self, validator_tag:str, key:Key, value:str, references:dict) -> ValidatorOutput:
        """
        Validates a value from a model

        Using the validator_tag parameter and config files, a validator is invoked that validates the value.  
        Success or failure is written as a property of the 'key' parameter under the validator name e.g.
        'validators.date'

        Parameters
        ----------
        validator_tag : str
            The tag that will be looked up in the validator config YAML to get the validator method and config to use
        key : data.key.Key
            The key from the model that had the tag.  Mostly used to report errors.
        value : str
            The value to validate
        references : dict
            A dict that should contain at least a 'template-model' key that contains the 
            template model.  May be customised to contain anything though.

        Returns
        -------
        ValidatorOutput class
        """

        output = ValidatorOutput(validator_tag, key, value)

        # Check the validator tag can be found
        if (validator_entry := self.validator_config_dict.get(validator_tag)) is None:
            logger.error(f"No validator tag '{validator_tag}' is configured")
            output.validator_result = False
            output.error = f"No validator tag '{validator_tag}' is configured"
            return output

        # Check the validator is present
        if validator_entry.get('validator') is None:
            logger.error(f"No 'validator' key is configured for '{validator_tag}'")
            output.validator_result = False
            output.error = f"No 'validator' key is configured for '{validator_tag}'"
            return output

        # Get the validator
        if (validate := self.dispatch.get(validator_entry.get('validator'))) is None:
            logger.error(f"No validator called '{validator_entry['validator']}' is configured")
            output.validator_result = False
            output.error = f"No validator called '{validator_entry['validator']}' is configured"
            return output

        references['validator-tag'] = validator_tag

        # Validate
        #logger.info(f"Entering validator '{validator_entry['validator']}'")
        output.validator_result = validate(validator_entry['config'], key, value, references)
        #logger.info(f"Exiting validator '{validator_entry['validator']}'")
        
        output.validator_name = validator_entry["validator"]
        output.validator_module = self.modules[validator_entry["validator"]]
        output.description = key.getProperty(validator_tag)

        return output


    def validate_from_config(self, validator_config:dict, key:Key, value:str, references:dict) -> bool:
        """
        Validates a Key value

        Parameters
        ----------
        validator_name : str
            Name of the validator to invoke
        config : dict
            The configuration for the validator
        key : data.key.Key
            The key from the model that had the tag.  Mostly used to report errors.
        value : str
            The value to validate
        references : dict
            A dict that should contain at least a 'template-model' key that contains the 
            template model.  May be customised to contain anything though.

        Returns
        -------
        bool : Did the value validate? True if it did, False otherwise
        """

        validator_name = validator_config.get("validator", "")
        config = validator_config.get("config", {})

        # Get the validator
        if (validate := self.dispatch.get(validator_name)) is None:
            logger.error(f"No validator called '{validator_name}' is configured")
            return False

        # Validate
        result = validate(config, key, value, references)
        
        return result



#!/usr/bin/env python3
"""
Manage a threat model
"""

import logging
from datetime import datetime
from dateutil.parser import parse
from utils.output import FormatOutput
from utils.error import ManageError, StorageError
from utils.model import assign_row_identifiers, assign_parent_keys
from data import find
from utils import match
from storage.gitrepo import GitStorage
from manage import manage_config
from manage.metadata import IndexMetaData, ThreatModelMetaData
from manage.metadata import MetadataIndexEntry, ThreatModelVersionMetaData
from manage.manage_storage import IndexStorage
from manage.manage_storage import ThreatModelStorage

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def config():

    return manage_config.config()

def _output(config:dict):

    return FormatOutput(config.get("output", {}))


def indexdata(config:dict, execution_env, ID:str):
    """ Given a document ID retrieve the metadata associated with the threat model """

    output = _output(config)

    try: 
        if ID is None:
            logger.error(f"An ID must be provided")
            raise ManageError("internal-error", {})

        with GitStorage(config.get("storage", {}), execution_env) as storage:
        
            index = IndexMetaData(config.get("metadata", {}), storage)
            
            entry = index.getIndexEntry(ID)
            if entry is None:
                logger.error(f"An entry with ID '{ID}' was not found")
                raise ManageError("no-index-ID", {"ID":ID})

            output.setSuccess("success-indexdata", {"ID":ID}, entry.get_state())
    
    except StorageError as error:
        output.setError(error.text_key, error.template_values)
    except ManageError as error:
        output.setError(error.text_key, error.template_values)

    return output


def create(config:dict, execution_env, IDprefix:str, scheme:str, location:str, model:dict):
    """ Given a document location and id format, generate the unique document ID, and store and return the result """

    output = _output(config)

    entry_exists = False
    try:
        assign_row_identifiers(model)

        if IDprefix is None or scheme is None or location is None:
            logger.error(f"An ID prefix, scheme and location, must all be provided")
            raise ManageError("internal-error", {})

        with IndexStorage(config.get("storage", {}), execution_env, output_texts=output.templated_texts) as storage:

            index = IndexMetaData(config.get("metadata", {}), storage)

            entry = index.getIndexEntryByLocation(location)
            if entry is not None:
                entry_exists = True
                # Need to tell IndexStorage to not try to persist anything, because there has been no changes.
                storage.persistChanges(False)
            else:
                new_ID = index.createIndexEntryID(IDprefix)

                new_entry = MetadataIndexEntry(config.get("metadata", {}), {})
                new_entry.createCurrentVersionEntryFromModel(scheme, location, model)
                new_entry.ID = new_ID
                # We never create new metadata for a threat model indicating it is approved.  Only Submit can change the status
                new_entry.stripApproval()

                index.setIndexEntry(new_ID, new_entry)

                index.persist()

        if entry_exists:
            # In the rare case the entry already exists, we determine if existing extry is on default branch or not, 
            # so users aren't confused why they can't see the TM ID if they call indexdata (which only returns approved TM metadata)
            indexdata_output = indexdata(config, execution_env, entry.ID)
            default_existing_entry = indexdata_output.getDetails()

            if default_existing_entry is not None:
                # Existing entry is on default branch
                logger.info(f"Cannot create ID for threat model in location '{location}' as that location is used by threat model '{entry.ID}'")
                raise ManageError("index-entry-exists", {"location":location, "ID":entry.ID})
            else:
                # Existing entry is not on default branch
                logger.info(f"Cannot create ID for threat model in location '{location}' as that location is used by threat model '{entry.ID}', but this ID allocation has not been merged yet")
                raise ManageError("index-entry-exists-create", {"location":location, "ID":entry.ID})

        output.setSuccess("success-create", {}, new_entry.get_state())

    except StorageError as error:
        output.setError(error.text_key, error.template_values)
    except ManageError as error:
        output.setError(error.text_key, error.template_values)

    return output


def submit(config:dict, execution_env, location:str, schemeID:str, model:dict):
    """ 
    Submit the threat model for approval 
    
    Given config and a threat model ID, check to see this TM has been submitted before, and if not create dir.  Otherwise, get the TM, get the version,
    and metadata, and create/update metadata for that version.
    """

    output = _output(config)

    try:
        assign_row_identifiers(model)

        # Parse for the index metadata to store
        imdCurrent = MetadataIndexEntry(config.get("metadata", {}), {})
        current_version = imdCurrent.createCurrentVersionEntryFromModel(schemeID, location, model)

        if match.is_empty(imdCurrent.ID):
            raise ManageError("empty-id", {})

        imdApproved = MetadataIndexEntry(config.get("metadata", {}), {})
        approved_version = imdApproved.createApprovedVersionEntryFromModel(schemeID, location, model)

        # Parse for complete metadata to store
        tmvmd = ThreatModelVersionMetaData()
        tmvmd.fromModel(schemeID, location, model)

        with ThreatModelStorage(config.get("storage", {}), execution_env, imdCurrent.ID, persist_changes=True, output_texts=output.templated_texts) as storage:

            tm_metadata = ThreatModelMetaData(config.get("metadata", {}), storage, schemeID, location)

            if not match.equals(imdCurrent.ID, tm_metadata.ID):
                logger.error(f"The ID value in the threat model ('{imdCurrent.ID}') does not match the ID value in the metadata ('{tm_metadata.ID}') for the registered scheme and location")
                raise ManageError("model-and-metadata-ID-mismatch", {"current_ID":imdCurrent.ID, "metadata_ID":tm_metadata.ID})

            # Add/update the metadata related to this version
            tm_metadata.setApprovedVersion(imdCurrent.ID, current_version, imdCurrent, approved_version, imdApproved, tmvmd)

            # The model will be recorded as well
            tm_metadata.setModel(approved_version, model)

            tm_metadata.persist()

        if match.equals(current_version, approved_version):
            output.setSuccess("success-approver", {"ID":imdCurrent.ID}, tmvmd)
        else:
            output.setSuccess("success-submitter", {"ID":imdCurrent.ID}, tmvmd)

    except StorageError as error:
        output.setError(error.text_key, error.template_values)
    except ManageError as error:
        output.setError(error.text_key, error.template_values)

    return output

def check(config:dict, execution_env, location:str, schemeID:str, model:dict, measure_config:dict, distance):
    """ Given a document, check if the threat model has changed enough from the approved version as to require re-approval """

    output = _output(config)

    approval_expiry_days = config["check"]["approval-expiry-days"]

    try:
        assign_row_identifiers(model)
        assign_parent_keys(model)

        # Get the Doc ID
        if (docIDtuple := find.key_with_tag(model, "document-id")) is None:
            logger.error(f"No document ID was found for the threat model at location {location}")
            raise ManageError("no-document-id", {"location":location})
        _, docID = docIDtuple

        with ThreatModelStorage(config.get("storage", {}), execution_env, docID, persist_changes=False) as storage:

            # Get the index data and metadata for the ID
            tm_metadata = ThreatModelMetaData(config.get("metadata", {}), storage, schemeID, location)

            if not match.equals(docID, tm_metadata.ID):
                logger.error(f"The ID value in the threat model does not match the ID value in the metadata for the registered scheme and location")
                raise ManageError("model-and-metadata-ID-mismatch", {})

            indexentry = tm_metadata.index.getIndexEntryByLocation(schemeID, location)

            # Check if there is an approved version
            if indexentry is None or (approved_version := indexentry.approved_version) is None or indexentry.approved_date is None:
                # Need an approval
                raise ManageError("no-approved-version", {})

            # Check if approved version has expired
            if indexentry.approved_date is not None:
                approvedDate = parse(indexentry.approved_date)
                if approvedDate is not None and (datetime.now() - approvedDate).days > approval_expiry_days:
                    raise ManageError("approved-version-expired", {"approval_expiry_days":approval_expiry_days})

            # Get approved version
            if (approved_model := tm_metadata.load_model(approved_version)) is None:
                raise ManageError("approved-version-didnt-load", {})
                      
            # Potentially need to add 'measure' tags in the right place
            
            # Use measure code to find things that don't match between a current TM and the most recent approved version
            #measure_config = measure.config(config.translator)
            # Update the tag prefix to not collide with any other measures configured
            measure_config["measure-tag"]["prefix"] = "check"
            #measure_output = measure.output(measure_config)
            #measure.distance(measure_config, measure_output, model, approved_model)
            measure_output = distance(measure_config, model, approved_model)

            # Are there new threats in the current TM
            # Are there new controls in the current TM
            meaure_metric = measure_output.get_measure_metric()
            if meaure_metric != "0%":
                # An approval is required
                output.setInformation("approval-required", {}, measure_output.getMeasureDetails())
            else:
                # An approval is not required
                output.setInformation("no-approval-required", {})

    except StorageError as error:
        output.setError(error.text_key, error.template_values)
    except ManageError as error:
        output.setError(error.text_key, error.template_values)

    return output
    

    

    
    
    

    


#!/usr/bin/env python3
"""
Verify a threat model
"""

import logging
from utils.error import VerifyError
from utils.output import FormatOutput
from utils import match
from verifiers.verifiers_config import VerifiersConfig
from verifiers.threat_coverage import ThreatCoverage
from verifiers.verifiers_report import VerifiersReport
from verifiers.verifiers import Verifiers

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def config(scheme:dict):

    verifiers_config = scheme.get("verifiers", {})

    return VerifiersConfig(verifiers_config)


def assign_default_tags(config:VerifiersConfig, model:dict):
    """
    Utility method to populate a model with the default verify tags.
    """
    verifiers = Verifiers(config)

    verifiers.assign_key_tags(model)


def verify(config:VerifiersConfig, threatmodel:dict, tm_template:dict):

    logger.info("Entering verify")

    output = FormatOutput(config.verifiers_config_dict.get("output"))

    try: 
        verifiers = Verifiers(config)

        issues = verifiers.verify(threatmodel, tm_template)

        if len(issues) == 0:
            output.setSuccess("success-no-issues", {}, issues)
        elif len([issue for issue in issues if issue.isError()]) > 0:
            output.setInformation("error-issues", {}, issues)
        else:
            output.setInformation("warn-info-issues", {}, issues)

    except VerifyError as error:
        output.setError(error.text_key, error.template_values)

    logger.info("Exiting verify")

    return output

def _coverage(config:VerifiersConfig, threatmodel:dict):

    threat_coverage = ThreatCoverage(config)

    coverage = threat_coverage.coverage(threatmodel)

    return coverage

def _reports_to_show(reports_parameter:str):
    """ Returns a bool 2-tuple indicating if the asset and control reports should be shown """

    if match.is_empty(reports_parameter):
        return False, False

    if reports_parameter == "none":
        return False, False
    elif reports_parameter == "assets":
        return True, False
    elif reports_parameter == "controls":
        return False, True
    elif reports_parameter == "all":
        return True, True

    logger.warning(f"Unrecognised value '{reports_parameter}' for 'reports' parameter.  Using default.")
    return False, False

def report(config:VerifiersConfig, threatmodel:dict, issues:list, verify_reports:str):

    logger.info("Entering report")

    output = FormatOutput(config.verifiers_config_dict.get("output"))

    try:
        
        coverage = _coverage(config, threatmodel)

        show_asset_report, show_control_report = _reports_to_show(verify_reports)

        verifiers_report = VerifiersReport(config, show_asset_report, show_control_report)

        verifiers_report.report(issues, coverage)

        if len(issues) == 0:
            output.setSuccess("success-no-issues", {}, verifiers_report)
        elif len([issue for issue in issues if issue.isError()]) > 0:
            output.setInformation("error-issues", {}, verifiers_report)
        else:
            output.setInformation("warn-info-issues", {}, verifiers_report)        

    except VerifyError as error:
        output.setError(error.text_key, error.template_values)

    logger.info("Exiting report")

    return output



#!/usr/bin/env python3
"""
Makes available configuration for AWS Lambda execution context
"""

import logging
import os
import json
from sh import ssh_keyscan
from sh.contrib import git
from pathlib import Path
import utils.shell as shell
from utils.error import StorageError
import boto3
from botocore.exceptions import ClientError

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class AWSLambdaContext:

    def __init__(self, config:dict):
        # Needs to be able to handle an empty config because providers may get called before config files are available.

        self.config = config
        if "secret_name" in config:
            self.secret_name = config.get("secret_name")
        else:
            # Read from env var if possible
            self.secret_name = os.getenv("THREATWARE_AWS_SECRET_NAME")

        if "region" in config: 
            self.region = config.get("region")
        else:
            # Read from env var if possible
            self.region = os.getenv("THREATWARE_AWS_SECRET_REGION")

        self.secret_dict = {}
        if self.secret_name is not None and self.region is not None:
            self.secret_dict = json.loads(self._get_secret())

    def _get_secret(self):

        # This code taken from https://boto3.amazonaws.com/v1/documentation/api/latest/guide/secrets-manager.html

        secret_name = self.secret_name
        region_name = self.region

        session = boto3.session.Session()
        client = session.client(
            service_name='secretsmanager',
            region_name=region_name,
        )

        try:
            get_secret_value_response = client.get_secret_value(
                SecretId=secret_name
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                logger.error(f"The requested secret '{secret_name}' was not found")
            elif e.response['Error']['Code'] == 'InvalidRequestException':
                logger.error(f"The request was invalid due to: {e}")
            elif e.response['Error']['Code'] == 'InvalidParameterException':
                logger.error(f"The request had invalid params: {e}")
            elif e.response['Error']['Code'] == 'DecryptionFailure':
                logger.error(f"The requested secret can't be decrypted using the provided KMS key: {e}")
            elif e.response['Error']['Code'] == 'InternalServiceError':
                logger.error(f"An error occurred on service side: {e}")
        else:
            # Secrets Manager decrypts the secret value using the associated KMS CMK
            # Depending on whether the secret was a string or binary, only one of these fields will be populated
            if 'SecretString' in get_secret_value_response:
                text_secret_data = get_secret_value_response['SecretString']
            else:
                logger.error(f"No 'SecretString' was available in secret '{secret_name}'")
                #binary_secret_data = get_secret_value_response['SecretBinary']

        return text_secret_data

    def getConfluenceConnectionCredentials(self):
        
        confluence_conn = json.loads(self.secret_dict.get("confluence"))

        return confluence_conn

    def getGoogleCredentials(self):

        credentials = json.loads(self.secret_dict.get("google"))

        return credentials

    def getGitCredentials(self):

        credentials = json.loads(self.secret_dict.get("git"))

        return credentials
    
    def _format_openssh_private_key(self, key:str) -> str:
        """
        Reformat OpenSSH private key in-case user has just cut and paste

        Ensures OpenSSH header and footer text ends in newline - without this ssh considers the private key file invalid
        """
        openssh_header = "-----BEGIN OPENSSH PRIVATE KEY-----"
        openssh_footer = "-----END OPENSSH PRIVATE KEY-----"

        # Check if key contains correct newline terminated header
        if key.find(openssh_header + "\n") == -1:
            # Check if key contains header, just not newline terminated
            if key.find(openssh_header) != -1:
                # Replace non newline terminated header with newline terminated header
                key = key.replace(openssh_header, openssh_header + "\n")
            else:
                # Not an OpenSSH private key
                return key

        # Check if key contains correct newline terminated footer
        if key.find(openssh_footer + "\n") == -1:
            # Check if key contains footer, just not newline terminated
            if key.find(openssh_footer) != -1:
                # Replace non newline terminated footer with newline terminated footer
                key = key.replace(openssh_footer, openssh_footer + "\n")
            else:
                logger.error("Private does not have an OpenSSH footer")
                # Not an OpenSSH private key
                return key

        return key

    def setupGit(self, remote:str, base_storage_dir:str):

        # On AWS Lambda, to setup git, since we can't write to the local user's /home directory, we must
        # configure git in the only location we can write to, which is /tmp

        if remote.startswith("git@"):
            logger.info(f"Setting up git to use SSH credentials in {base_storage_dir}")
            ssh_path = Path.joinpath(Path(base_storage_dir), ".ssh")
            ssh_config_path = str(Path.joinpath(ssh_path, "config"))
            ssh_private_key_path = str(Path.joinpath(ssh_path, "key"))
            ssh_public_key_path = str(Path.joinpath(ssh_path, "key.pub"))
            ssh_known_hosts_path = str(Path.joinpath(ssh_path, "known_hosts"))
            ssh_path = str(ssh_path)

            # To work in lambda which doesn't have '/dev/fd' (which this argument that defaults to true relies on), we need to set this to False.  Hope this has no negative side effects
            shell.global_kwargs["_close_fds"] = False
            # To work in lambda, which seems to have no or few TTYs (because you get errors about it running out of them), set this to False
            shell.global_kwargs["_tty_out"] = False

            # Change the location where git config looks for the config file
            os.makedirs(os.path.join(base_storage_dir, "git"), exist_ok=True)
            #with open(os.path.join(base_storage_dir, "git", "config"), mode='a'): pass
            # Other options are to set
            # GIT_CONFIG_GLOBAL to the name of the global config file to use
            # GIT_CONFIG_SYSTEM to /dev/null to prevent git from reading the system config file (or set GIT_CONFIG_NOSYSTEM to 1)

            # To work in lambda which only allows us to write to /tmp, we need to configure ssh to use the config file there, and we need to set this env var so git uses ssh with this file
            git_env = os.environ.copy()
            git_env["GIT_SSH_COMMAND"] = "ssh -F " + ssh_config_path
            # git config complains if HOME is not set
            git_env["HOME"] = base_storage_dir
            git_env["XDG_CONFIG_HOME"] = base_storage_dir
            shell.global_kwargs["_env"] = git_env

            # The dockerfile created a symlink to /tmp/.ssh, but in AWS lambda that doesn't exist, so we need to create it.
            os.makedirs(ssh_path, exist_ok=True)

            # Write private and public keys to /tmp/.ssh/key and /tmp/.ssh/key.pub
            git_keys = self.getGitCredentials()
            if git_keys is None or len(git_keys) == 0:
                raise StorageError("storage.no-git-credentials", {})
            git_private_key = git_keys["private-key"]
            git_public_key = git_keys["public-key"]
            git_public_key_list = git_public_key.split(" ")
            if len(git_public_key_list) != 3:
                logger.error(f"Expected public key to consist of 3 space separated parts, instead found {len(git_public_key_list)} parts")
                raise StorageError("internal-error", None)
            git_alg, git_key, git_email = git_public_key_list[0], git_public_key_list[1], git_public_key_list[2]

            # ssh is REALLY fussy about the private key file contents, so we need to do a sanity check on the contents we read in and ensure it will parse correctly
            # Header and footer lines need to be be perfect i.e. "-----BEGIN OPENSSH PRIVATE KEY-----\n", "-----END OPENSSH PRIVATE KEY-----\n"
            git_private_key = self._format_openssh_private_key(git_private_key)

            with open(os.open(ssh_private_key_path, os.O_CREAT | os.O_WRONLY, 0o600), "w") as key_file:
                key_file.writelines(git_private_key)
                # Private key must have restricted access or git complains.
            with open(ssh_public_key_path, "w") as key_file:
                key_file.writelines(git_public_key)

            # Write .ssh/config for the key file
            with open(ssh_config_path, "w") as ssh_config:
                ssh_config.writelines(["IdentityFile=" + ssh_private_key_path + "\n", "UserKnownHostsFile=" + ssh_known_hosts_path + "\n"])

            # Get the host from the 'remote' value of the form 'protocol@host:path'
            #if not self.remote.startswith("git@"):
            #    logger.error("Currently only remotes using 'git@' are supported")
            #    raise StorageError("internal-error", None)
            git_host = (remote.split("@")[1]).split(":")[0]

            # Need to populate .ssh/known_hosts with remote pub key i.e. ssh-keyscan -t ed25519 github.com >> /tmp/.ssh/known_hosts
            shell.run(base_storage_dir, ssh_keyscan, ["-t", git_alg, git_host], _out=ssh_known_hosts_path)

            # Git must know who the user is before it can commit. Configure git user.name and user.email
            shell.run(base_storage_dir, git.config, ["--global", "user.name", "threatware"])
            shell.run(base_storage_dir, git.config, ["--global", "user.email", "threatware"])

        elif remote.startswith("http"):
            logger.info("Using git with anonymous HTTP")
        else:
            logger.warning("Unrecognised git remote, assuming anonymous HTTP")

    def get_config_base_dir(self, suggested_base_dir):
        """ Return the directory where we expect the configuration file to be based 
        
        Returns
        -------
        str : The suggested base directory to find configuration file
        bool : Whether this environment is ephemeral (config will not persist between invocations)
        """

        # We can only write to /tmp for lambdas, so need to check suggestion does that, and if not move it there
        real_suggested_base_dir = os.path.realpath(os.path.normpath(suggested_base_dir))

        # Compare dir /tmp to the shared path between /tmp and the suggestion (which will be /tmp if the suggested path starts with /tmp)
        if os.path.samefile("/tmp", os.path.commonpath(["/tmp", real_suggested_base_dir])):
            base_dir = real_suggested_base_dir
        else:
            if os.path.isabs(real_suggested_base_dir):
                head, tail = os.path.split(real_suggested_base_dir)
                base_dir = os.path.join("/tmp", tail)
            else:
                base_dir = os.path.join("/tmp", real_suggested_base_dir)

        # The config will NOT persist after execution
        return base_dir, True

def load(config:dict):

    return AWSLambdaContext(config)


#!/usr/bin/env python3

import logging
import data.value
import data.output
from . import key
from utils.property_str import pstr

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def parse(map_data_def, input) -> dict:
    """ Traverses a map-data: definition and calls value.parse on each value """

    logger.debug(f'Entering: Key count = {len(map_data_def)}, data count = {len(input) if isinstance(input, list) else 1}')

    output = {}

    for key_value in map_data_def:
        if isinstance(key_value["key"], str):
            key_def = key.key(key_value["key"])
        elif isinstance(key_value["key"], dict):
            if (key_def := key.key(key_value["key"]["name"])) is None:
                # TODO throw exception or handle
                logger.error(f"The key {key_value['key']} must have a 'name' key with a non-empty value")
            if (section := key_value["key"].get("section")) is not None:
                # Record a friendly section name for this key, if one exists e.g. a table name
                key_def.addProperty("section", section)
            if (tags := key_value["key"].get("tags")) is not None and isinstance(tags, list):
                key_def.addTags(tags)


        logger.debug(f"Mapping data to '{key_def}'")

        value_def = key_value["value"]
        value = data.value.parse(value_def, input)
        
        logger.debug(f"Mapping '{key_def}':'{value}'")

        # We have a value to map to the key, but it might not be a simple value, or it might be an already processed value mapping to 
        # a key up in the hierarchy.  Need to figure out if it's a value with additional information we need to store against the key.
        if isinstance(value, pstr):
            for dict_key, dict_value in value.properties.items():
                key_def.addProperty(dict_key, dict_value)
            output[key_def] = value.to_str()
        else:
            output[key_def] = value

        if key_def.getProperty("section") is not None:
            # Let the key reference the value for these high level sections. Makes searching sections e.g. tables, easier given only a tagged key.
            if isinstance(value, pstr):
                key_def.addProperty("value", value.to_str())
            else:
                key_def.addProperty("value", value)

    logger.debug(f'Leaving: Key count = {len(map_data_def)}, data count = {len(input) if isinstance(input, list) else 1}')

    return output


#!/usr/bin/env python3
"""
Converts a Threat Model document into a model
"""

import logging
from utils.error import ConvertError, ProviderError
from utils.output import FormatOutput
from convertors import convertors_config
import convertors.confluence_convertor.convertor
import convertors.gdoc_convertor.convertor
from utils.output import FormatOutput

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def config():

    return convertors_config.config()


def _output(config:dict):

    return FormatOutput(config.get("output", {}))


def convert(config:dict, execution_env, scheme:dict, doc_location:str, store_doc:bool=True):

    logger.info("Entering convert")
    
    output = _output(config)

    try: 
        if scheme['document-storage'] == "confluence":
            model = convertors.confluence_convertor.convertor.convert(config, execution_env.getConfluenceConnectionCredentials(), scheme, {"id":doc_location}, store_doc)
        elif scheme['document-storage'] == "googledoc":
            model = convertors.gdoc_convertor.convertor.convert(config, execution_env.getGoogleCredentials(), scheme, {"id":doc_location}, store_doc)
        else:
            logger.error(f"Unknown document storage type '{scheme['document-storage']}'")
            raise ConvertError("unknown-doc-storage", {"doc_storage":scheme['document-storage']})

        output.setSuccess("success", {}, model)

    except (ConvertError, ProviderError) as error:
        output.setError(error.text_key, error.template_values)

    logger.info("Exiting convert")

    return output


def convert_template(config:dict, execution_env, scheme:dict, doc_location:str):

    logger.info("Entering convert_template")

    output = _output(config)

    try: 

        if not doc_location:
            doc_location = scheme['template-id']
            if not doc_location:
                logger.error("No template document location provided as input or defined in the scheme")
        
        return convert(config, execution_env, scheme, doc_location, store_doc=False)
    
    except ConvertError as error:
        output.setError(error.text_key, error.template_values)

    logger.info("Exiting convert_template")

    return output




#!/usr/bin/env python3
"""
Class to download a git repo
"""
import logging
from io import StringIO
from storage.gitrepo import GitStorage
import utils.match as match
import utils.shell as shell
from sh.contrib import git
from utils.error import StorageError

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class GitPermanentStorage(GitStorage):

    is_entered = True

    def __init__(self, config:dict, execution_env):

        super().__init__(config, execution_env)

        self.is_entered = True

    def clone(self, target_dir:str, branch:str):

        buf = StringIO()
        
        #if Path(self.repodir).is_dir():
        #    logger.warning(f"Directory '{self.repodir}' already exists")

        if match.is_empty(branch):
            git_args = ["--depth", "1", self.remote, target_dir]
        else:
            git_args = ["--depth", "1", "--branch", branch, self.remote, target_dir]

        if not shell.run(self.base_storage_dir, git.clone, git_args, _out=buf, _err_to_out=True):
            outdata = buf.getvalue()
            logger.error(f"Could not clone repo '{self.remote}'.  Error message: {outdata}")
            raise StorageError("storage.clone-error", {"storage":{"remote":self.remote}})




#!/usr/bin/env python3
"""
Loads YAML
"""

import logging
from io import StringIO
from utils.error import ThreatwareError
from ruamel.yaml import YAML

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

##
## Methods to help with reading yaml
##

def yaml_file_to_dict(path:str, output_classes:list = []):

    yaml=YAML(typ='safe')   # default, if not specfied, is 'rt' (round-trip)
    yaml.default_flow_style = False
    logger.debug("Loading YAML '{}'".format(path))

    try: 
        file = open(path, 'r')
    except OSError:
        raise ThreatwareError

    #with open(str(Path(__file__).absolute().parent.joinpath(path)), 'r') as file:
    #with open(path, 'r') as file:
    for output_class in output_classes:
        yaml.register_class(output_class)
    
    with file:
        yamldict = yaml.load(file)
        return yamldict

def yaml_file_to_str(path:str):

    try: 
        file = open(path, 'r')
    except OSError:
        raise ThreatwareError

    with file:
        yamlstr = file.read()
        return yamlstr

def yaml_str_to_dict(yamlstr:str):

    yaml=YAML(typ='safe') 
    yamldict = yaml.load(yamlstr)
    return yamldict


##
## Methods to help with outputting yaml files
##

_registered_classes:list = []

def yaml_register_class(class_type):
    """ To output to yaml we need to register classes that do their own YAML serialisation. """
    global _registered_classes
    
    if class_type not in _registered_classes:
        _registered_classes.append(class_type)

def _get_yaml_object():

    yaml=YAML(typ='safe')
    yaml.default_flow_style = False
    yaml.indent = 4
    yaml.sort_base_mapping_type_on_output = False
    for registered_class in _registered_classes:
        yaml.register_class(registered_class)

    return yaml

def class_to_yaml_file(output_instance, file_location):

    yaml = _get_yaml_object()

    yaml.dump(output_instance, file_location)


def class_to_yaml_str(output_instance) -> str:
    """ Output an object in YAML.  Any classes in the object need to be registered using yaml_register_class (and provide the appropriate serialisation methods). """
    
    yaml = _get_yaml_object()

    with StringIO() as buf:
        yaml.dump(output_instance, buf)

        return buf.getvalue()


#!/usr/bin/env python3
"""
Configuration loading for Manage action
"""

import logging
from pathlib import Path
from utils.config import ConfigBase
from language.translate import Translate
import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

MANAGE_CONFIG_YAML = "manage_config.yaml"
MANAGE_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(MANAGE_CONFIG_YAML))

def config() -> dict:

    yaml_config_dict = Translate.localiseYamlFile(ConfigBase.getConfigPath(MANAGE_CONFIG_YAML_PATH))

    return yaml_config_dict["manage-config"]


#!/usr/bin/env python3
"""
Run shell commands
"""

import logging
from sh import pushd
from sh import ErrorReturnCode

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

global_kwargs = {}

def run(directory, command, *args, **kwargs):
    """
    Run a shell command using the sh package

    Generically runs a sh command, with logging, and captures output as specified

    Parameters
    ----------
    directory : str
        The directory where to execute the command
    command 
        The sh command to execute
    *args : list
        The args to pass to the command
    **kwargs : list
        The named args to pass to the command.  Use "_out" named argument to specify output (se sh package)

    Returns
    -------
    bool : True if no errors occurred, False otherwise
    """

    if "_out" not in kwargs:
        kwargs["_out"] = logger.debug

    command_kwargs = global_kwargs | kwargs

    with pushd(directory):  
        try:
            if logger.getEffectiveLevel() == logging.DEBUG:
                # Don't print out env vars
                debug_kwargs = {key: value for key, value in command_kwargs.items() if key != "_env"}
                logger.debug(f"Running '{command.__name__}' with args '{args}' and kwargs '{debug_kwargs}' in directory {directory}")
            command(*args, **command_kwargs)
        except ErrorReturnCode as erc:
            try:
                stdout_data = erc.stdout.decode()
                stderr_data = erc.stderr.decode()
            except (UnicodeDecodeError, AttributeError):
                pass

            logger.error(f"'{erc.full_cmd}' failed with exit code '{erc.exit_code}' and stdout '{stdout_data}' and stderr '{stderr_data}'")
            return False

    return True        

    


#!/usr/bin/env python3
"""
Class Response
"""

import logging
from utils.config import ConfigBase
from utils.output import FormatOutput
from utils.load_yaml import yaml_file_to_dict
from data.key import key as Key
from language.translate import Translate
from utils.request import Request
from response.response_config import ResponseConfig
from response.html_response import get_html_response

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

#OUTPUT_TEXTS_YAML = "output_texts.yaml"
#OUTPUT_TEXTS_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(OUTPUT_TEXTS_YAML))

class Response:

    html_document:str = ""

    def __init__(self, output:FormatOutput, meta_override:str = None, force_api_format:bool = False):

        self.format = Request.format
        if force_api_format:
            # We likely have an error situation and don't want to try and return html
            if self.format != "json" or self.format != "yaml":
                self.format = "json"
        self.output = output
        self.meta_override = meta_override

        self.response_config = ResponseConfig()
        self.templated_texts = yaml_file_to_dict(ConfigBase.getConfigPath(self.response_config.template_text_file)).get("output-texts")

    
    def getContentType(self):

        if self.format == "yaml":
            return "text/yaml"
        if self.format == "html":
            return "text/html"
        
        return "application/json"

    def getBody(self):

        if self.meta_override is not None:
            Key.config_serialisation(self.meta_override)

        if self.format == "yaml":
            return self.output.toyaml()
        if self.format == "html":
            return get_html_response(self.response_config.htmlConfig, self.output, self.html_document, self.templated_texts)
        
        return self.output.tojson()

    @classmethod
    def setDocument(cls, document):

        cls.html_document = document


#!/usr/bin/env python3
"""
Custom Exception
"""

class ThreatwareError(Exception):

    def __init__(self, text_key:str = "", template_values:dict = {}, *args: object) -> None:
        self.text_key = "internal-error"
        if text_key is not None and len(text_key.strip()) > 0:
            self.text_key = text_key
        self.template_values = template_values
        super().__init__(*args)

class SchemeError(ThreatwareError):
    pass

class ConvertError(ThreatwareError):
    pass

class ProviderError(ConvertError):
    pass

class ValidatorsError(ThreatwareError):
    pass

class VerifyError(ThreatwareError):
    pass

class ManageError(ThreatwareError):
    pass
    # def __init__(self, text_key:str, template_values:dict, *args: object) -> None:
    #     self.text_key = text_key
    #     self.template_values = template_values
    #     super().__init__(*args)

class MeasureError(ThreatwareError):
    pass
    # def __init__(self, text_key:str, template_values:dict, *args: object) -> None:
    #     self.text_key = text_key
    #     self.template_values = template_values
    #     super().__init__(*args)

class StorageError(ThreatwareError):
    pass
    # def __init__(self, text_key:str, template_values:dict, *args: object) -> None:
    #     self.text_key = text_key
    #     self.template_values = template_values
    #     super().__init__(*args)


import logging
import jsonpickle, json
from typing import List
from lxml import etree
import lxml.html
from lxml.etree import XPathError
from utils.request import Request
from response.response_config import HTMLConfig, SchemeSpecificConfig, FindingsConfig, BannerConfig, HTMLInject
from utils.output import FormatOutput
from language.translate import Translate

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

THREATWARE_RESULT_ERRORS = "errors"
THREATWARE_RESULT_NO_ERRORS = "no-errors"

def _inject_into_HTML(document, config:HTMLInject):

    try:
        elements = document.xpath(config.location_element)
    except XPathError:
        logger.warning(f"XPath query '{config.location_element}' caused an error")
        
    if len(elements) == 0:
        logger.warning(f"XPath query '{config.location_element}' returned no elements")
        return
    if len(elements) > 1:
        logger.warning(f"XPath query '{config.location_element}' returned multiple elements, using first")
        
    element = elements[0]

    element_index = config.location_index

    element_list = []

    # Create the script includes
    for script in config.include_scripts:
        element_list.append(etree.Element("script", src=script))

    # Create the style includes
    for style in config.include_stylesheets:
        element_list.append(etree.Element("link", rel="stylesheet", href=style))
    
    # Create the inline element
    if config.inline_element:
        ele = etree.fromstring(config.inline_element)
        element_list.append(ele)                 

    # Create the inline style
    if config.inline_style:
        ele = etree.Element("style", type="text/css")
        ele.text = config.inline_style
        element_list.append(ele)

    # Create the inline script
    if config.inline_script:
        ele = etree.Element("script")
        ele.text = config.inline_script
        element_list.append(ele)

    # Insert the elements into the document
    if -1 == element_index:
        for sub_element in element_list:
            element.append(sub_element)
    else:
        for i, sub_element in enumerate(element_list):
            element.insert(element_index + i, sub_element)

def _inject_scheme_specfic_into_HTML(document, scheme_specific_config:List[SchemeSpecificConfig], scheme:str):

    for scheme_config in scheme_specific_config:
        if scheme_config.matches(scheme):
            _inject_into_HTML(document, scheme_config.htmlinject)


def _inject_banner_into_HTML(document, banner_config:BannerConfig, issues, showing_all_errors:bool, result:FormatOutput, template_values:dict):

    banner_msg = result.getDescription()
    threatware_result = THREATWARE_RESULT_NO_ERRORS
    if len(issues) > 0:
        threatware_result = THREATWARE_RESULT_ERRORS
        if not showing_all_errors:
            banner_msg = f"{banner_msg}{Translate.localise(template_values, banner_config.missingfindingstextkey)}"
            
    # Inject the banner to inform the user of the state of the action
    _inject_into_HTML(document, banner_config.htmlinject)

    # Inject the text we want to display.  Since we inject at position zero, this will run before the above banner config
    _inject_into_HTML(document, HTMLInject({
        "location": {
            "element": "//html/body", 
            "index": 0
        }, 
        "inline": {
            "script": f"""
                var banner_msg = \"{banner_msg}\";
                var threatware_result = \"{threatware_result}\";
            """
        }
    }))

def _inject_findings_into_HTML(document, issues:dict, findings_config:FindingsConfig, template_values:dict):
    """
    Injects the findings into the HTML document
    """
    showing_all_errors = True

    # Add attributes to the elements the findings are associated with
    finding_id_count = -1
    for issue in issues:
        try:
            finding_id_count += 1
            xpath = issue.issue_key.properties.get("location", "")
            if xpath == "":
                showing_all_errors = False
                continue
            element = document.xpath(xpath)[0]
            if findings_config.finding_attribute in element.attrib:
                element.attrib[findings_config.finding_attribute] = f"{element.attrib[findings_config.finding_attribute]} {finding_id_count}"
            else:
                element.attrib[findings_config.finding_attribute] = f"{finding_id_count}"

            if issue.errortype.name == "ERROR":
                element.attrib[findings_config.finding_type] = "error"
            elif issue.errortype.name == "WARNING" and findings_config.finding_type not in element.attrib:
                element.attrib[findings_config.finding_type] = "warning"
            elif issue.errortype.name == "INFO" and findings_config.finding_type not in element.attrib:
                element.attrib[findings_config.finding_type] = "info"

            if findings_config.finding_class in element.attrib:
                element.attrib[findings_config.finding_class] = f"{element.attrib[findings_config.finding_class]} {issue.verifier}"
            else:
                element.attrib[findings_config.finding_class] = f"{issue.verifier}"

        except XPathError:
            logging.warning(f"Could not find location {xpath} in the document")
        except IndexError:
            logging.warning(f"Could not find location {xpath} in the document")

    # Inject the findings into the document
    json_issues = jsonpickle.encode(issues, unpicklable=False)

    _inject_into_HTML(document, HTMLInject({
        "location": {
            "element": "//html/body", 
            "index": -1
        }, 
        "inline": {
            "script": f"""
                const threatwarefindings = {json_issues};
                const threatwareRequest = {json.dumps(Request.get())};
            """
        }
    }))

    # Inject the ability to display the findings
    _inject_into_HTML(document, findings_config.htmlinject)

    return showing_all_errors

def get_html_response(config:HTMLConfig, result:FormatOutput, html_doc:str, template_values:dict):

    #content = result.getDetails()
    #if "issues" not in content:
    #    logging.error(f"Expecting 'issues' key in content but none was found")
    issues = result.getDetails().getIssues()

    document = lxml.html.document_fromstring(html_doc)

    # Add any scheme specific content to make the exported HTML look like the original
    _inject_scheme_specfic_into_HTML(document, config.schemeSpecificConfig, Request.scheme)

    # Add the findings to the HTML document, and the ability to show the findings to the user
    showing_all_errors = _inject_findings_into_HTML(document, issues, config.findingsConfig, template_values)

    # Add a banner to the top of the body saying whether the validation passed, failed,or failed and not all errors are shown
    _inject_banner_into_HTML(document, config.bannerConfig, issues, showing_all_errors, result, template_values)

    new_html_bstr = lxml.html.tostring(document, encoding='unicode')
    new_html_str = new_html_bstr    # Required hack to get around the fact that the lxml.html.tostring() function returns a byte string, not a string

    # Add "<!doctype html>" to the top of the page.  This tells the brwoser to use standards mode.  This is important as the browser will use quirks mode if this is not present.
    new_html_str = "<!doctype html>\n" + new_html_str
    
    return new_html_str




#!/usr/bin/env python3
"""
Utility methods for data.key.Key
"""

import logging
import re
from data import find
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def get_section_for_key(row_key:Key) -> Key:
    if row_key is not None:
        sectionKey = row_key
        while (sectionKey.getProperty("section") is None) and (sectionKey.getProperty("parentKey") is not None):
            sectionKey = sectionKey.getProperty("parentKey")

        if sectionKey.getProperty("section") is not None:
            return sectionKey

    return None

def get_data_tag_for_key(row_key:Key) -> str:
    if row_key is not None:
        sectionKey = row_key
        
        while (len([tag for tag in sectionKey.getTags() if re.search("^.*-data$", tag)]) == 0) and (sectionKey.getProperty("parentKey") is not None):
            sectionKey = sectionKey.getProperty("parentKey")

        data_tag_list = [tag for tag in sectionKey.getTags() if re.search("^.*-data$", tag)]
        if len(data_tag_list) > 0:
            if len(data_tag_list) > 1:
                logger.warning(f"The key '{sectionKey}' should not be members of multiple data sections i.e. '{data_tag_list}'")
            return data_tag_list[0]

    return None

def get_row_identifier_for_key(row_key:Key):

    if row_key is not None:
        if row_key.hasTag("row-identifier"):
            return row_key
        elif row_key.getProperty("rowID") is not None:
            return row_key.getProperty("rowID")

    return None

def get_row_identifiers_for_key(row_key:Key):

    if row_key is not None:
        if row_key.hasTag("row-identifier") or row_key.getProperty("rowID") is not None:
            # It's possible the row_key is the same as the row identifier location, in which case the 
            # "rowID" property will not be set on it (it would be a circular reference)
            rowID_key = row_key.getProperty("rowID")
            if rowID_key is None:
                # row_key must have tag "row-identifier"
                return row_key.name, row_key.getProperty("value")
            else:
                return rowID_key.name, rowID_key.getProperty("value")

    return None, None

def get_column_name_for_key(row_key:Key) -> str:
    return row_key.getProperty("colname")

# def get_row_identifier_key(row_key:Key):

#     row_identifier_key, row_identifier_value = get_row_identifiers_for_key(row_key)

#     parentKeys = []
#     if row_identifier_key is None:
#         tableKey = row_key
#         # Navigate up the parents until we hit a table section
#         while (tableKey.getProperty("section") is None) and (tableKey.getProperty("parentKey") is not None):
#             parentKeys.append(tableKey)
#             tableKey = tableKey.getProperty("parentKey")

#         # Navigate back down, noting the row for the table
#         for row in tableKey.getProperty("value"):

#             def find_child_row(row, index):
#                 if isinstance(row, dict):
#                     row = [row]
#                 for row_entry in row:
#                     for row_key, row_value in row_entry.items():
#                         if row_key is parentKeys[index]:
#                             if index == 0:
#                                 return True
#                             return find_child_row(row_value, index - 1)
#                 return False
            
#             success = find_child_row(row, -1)
#             if success:
#                 row_identifier_key = find.key_with_tag(row, "row-identifier")
#                 return row_identifier_key
#     return None





#!/usr/bin/env python3
"""
Class ValidatorOutput
"""

from language.translate import Translate

import logging
from utils.load_yaml import yaml_register_class

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


class ValidatorOutput:
    """
    Holds the output of a validator and is used to format output of validator
    """

    templated_output_texts = {}
    validator_config = {}

    def __init__(self, validator_tag:str = None, validating_key = None, validating_value = None):
        """
        Create a ValidatorOutput object

        Parameters
        ----------
        validator_tag : str
            The tag that will be looked up in the validator config YAML to get the validator method and config to use
        validating_key : 
            The key of the entry being validated
        validating_value : 
            The value of the entry being validated
        """

        yaml_register_class(ValidatorOutput)

        self.validator_tag = validator_tag
        self.validating_key = validating_key
        self.validating_value = validating_value
        self.validator_name = None
        self.validator_module = None
        self.validator_result = False
        self.error = None

    def tag(self):
        return self.validator_tag
    
    def result(self):
        return self.validator_result
    
    def details(self):
        if not self.error:    
            context = {}
            context["key"] = {}
            context["key"]["name"] = self.validating_key.name
            context["key"]["colname"] = self.validating_key.getProperty("colname")
            context["key"]["value"] = self.validating_value
            context["config"] = ValidatorOutput.validator_config.get(self.validator_tag, {}).get("config")

            validator_output_texts = self.templated_output_texts.get(self.validator_tag, {}).get("text", {})
            if self.validator_result:
                return Translate.localise(validator_output_texts, "output_text_valid", context)
            else:
                return Translate.localise(validator_output_texts, "output_text_invalid", context)
        else:
            return self.error
        
    def _get_state(self):

        output = {}
        output['validator'] = self.tag()
        if not self.error:
            output['validates'] = self.result()
            output['description'] = self.details()
        else:
            output["error"] = self.error
        
        return output

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())


#!/usr/bin/env python3
"""
Class VerifierIssue
"""

import logging
from enum import Enum
from utils.load_yaml import yaml_register_class
from language.translate import Translate
from utils import keymaster
from utils.request import Request
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# from jinja2 import Environment, FileSystemLoader, select_autoescape
# env = Environment(
#     loader = FileSystemLoader(searchpath="./"),
#     autoescape=select_autoescape()
# )

class ErrorType(Enum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    NOT_SET = 4

    def __repr__(self):
        return self.name

class VerifierIssue:
    """
    Will output an issue with following format:
        type: ERROR
        table: The table where the issue exists
        table-row: The entry name and value where the issue exists
        error-description: A description of the error
        error-data: Any data related to the error
        fix-description: A description of a possible fix
        fix-data: Any data related to the fix
        verifier: The name of the verifier that detected the issue
    """

    # This will be set as a class variable so all Verifier instances use the same config
    issue_config = {"default-severity":"ERROR"}
    templated_error_texts = {}  # Will be read from file

    known_error_keys = [
        "issue_key",
        "issue_value",
        "issue_table",
        "errordata",
        "fixdata"
    ]

    def __init__(self, *, error_text_key:str, error_data_key:str = None, fix_text_key:str = None, fix_data_key:str = None, issue_dict:dict =None, errortype:ErrorType = ErrorType.NOT_SET):

        yaml_register_class(VerifierIssue)

        # given an 'issue_key' we want to construct a context so jinja can use e.g. {{ issuerow['component'].colname }} and {{ issuerow['component'].value }} to get, for instance, the information regarding the 'component' entry for a row.  Might not be that helpful for rows where some entries are multiple values, like for T&C 'components'


        self.verifier = ""  # Gets set externally
        self.errortype = errortype
        if errortype == ErrorType.NOT_SET:
            self.errortype = ErrorType[self.issue_config["default-severity"].upper()]

        if "issue_key" not in issue_dict:
            logger.error("The 'issue_dict' dict must include an 'issue_key' key with a Key object as it's value")
            self.error_desc = f"There was an problem generating the VerifierIssue"
            return
        self.issue_key = issue_dict["issue_key"]

        context = {}

        # Loop through all the values of issue_dict, and copy to context
        for keyentry, valueentry in issue_dict.items():
            # If any are Key objects, get name, colname, value, table, and create a row entry
            if isinstance(valueentry, Key):
                context[keyentry] = {}
                context[keyentry]["name"] = valueentry.name
                context[keyentry]["colname"] = valueentry.getProperty("colname")
                if context[keyentry]["colname"] is None:
                    context[keyentry]["colname"] = context[keyentry]["name"]
                context[keyentry]["value"] = issue_dict.get(keyentry.replace("_key", "_value"), None)
                if sectionKey := keymaster.get_section_for_key(valueentry):
                    context[keyentry]["table"] = sectionKey.getProperty("section")

                rowIDkey = keymaster.get_row_identifier_for_key(valueentry)
                if rowIDkey is not None:
                    row = rowIDkey.getProperty("row")
                    if row is not None:
                        rowentry = {}
                        for rowkeyentry, rowvalueentry in row.items():
                            rowentry[rowkeyentry.name] = {}
                            rowentry[rowkeyentry.name]["colname"] = rowkeyentry.getProperty("colname")
                            rowentry[rowkeyentry.name]["value"] = rowvalueentry
                            if sectionKey := keymaster.get_section_for_key(rowkeyentry):
                                rowentry[rowkeyentry.name]["table"] = sectionKey.getProperty("section")
                        context[keyentry]["row"] = rowentry
                    rowIDentry = {}
                    rowIDentry["name"] = rowIDkey.name
                    rowIDentry["colname"] = rowIDkey.getProperty("colname")
                    if rowIDentry["colname"] is None:
                        rowIDentry["colname"] = rowIDentry["name"]
                    rowIDentry["value"] = rowIDkey.getProperty("value")
                    if sectionKey := keymaster.get_section_for_key(rowIDkey):
                        rowIDentry["table"] = sectionKey.getProperty("section")
                    context[keyentry]["rowID"] = rowIDentry
                else:
                    # A 2 column table has no rowID
                    context[keyentry]["rowID"] = context[keyentry]
                    context[keyentry]["row"] = context[keyentry]
            else:
                context[keyentry] = valueentry

        # We support the caller setting the issue table to any value they want, but if not set, we populate it
        if context["issue_key"] is not None and context.get("issue_table") is None:
            sectionKey = keymaster.get_section_for_key(issue_dict["issue_key"])
            context["issue_table"] = sectionKey.getProperty("section")
        
        ### 'context' is set, so below here don't change it, just use it to render text ###
        
        # TODO gracefully handle templating exceptions from jinja, notably when it can't find the value when rendering

        # Set the issue table text
        self.issue_table_text = Translate.localise(self.templated_error_texts, "location-text", context)
        
        # We support the caller disabling the display of the issue table row information
        if "issue_table_row" in issue_dict and issue_dict.get("issue_table_row") is None:
            self.issue_table_row_text = None
        else:
            self.issue_table_row_text = Translate.localise(self.templated_error_texts, "entry-text", context)
        
        # Set the error description
        self.error_desc = Translate.localise(self.templated_error_texts, error_text_key, context)
        self.errordata = issue_dict.get("errordata")
        if error_data_key is not None and not Request.isAPIFormat():
            # For APIs (json/yaml) we want just the data, for other formats we want the data to be formatted
            self.errordata = Translate.localise(self.templated_error_texts, error_data_key, issue_dict)
        
        # Set any fix text specified
        if fix_text_key is not None:
            self.fix_desc = Translate.localise(self.templated_error_texts, fix_text_key, context)
        self.fixdata = issue_dict.get("fixdata")
        if fix_data_key is not None and not Request.isAPIFormat():
            # For APIs (json/yaml) we want just the data, for other formats we want the data to be formatted
            self.fixdata = Translate.localise(self.templated_error_texts, fix_data_key, issue_dict)

    def isError(self):
        return self.errortype == ErrorType.ERROR

    def isWarning(self):
        return self.errortype == ErrorType.WARNING

    def isInfo(self):
        return self.errortype == ErrorType.INFO

    def _get_state(self):

        # TODO - it would be good to substitute dict key names using the verifier_values.yaml file.  Probably need to recurse on things like fix-data

        # TODO - should probably move everything after the 'context' is set in __init__ to here, so we can late bind the output format to decide how to format the output

        output = {}
        output['type'] = self.errortype.name
        
        if getattr(self, "issue_table_text", None) is not None:
            output['table'] = self.issue_table_text
        if getattr(self, "issue_table_row_text", None) is not None:
            output['table-row'] = self.issue_table_row_text
        
        output['error-description'] = getattr(self, "error_desc", "")
        if getattr(self, "errordata", None) is not None:
            output['error-data'] = self.errordata

        if getattr(self, "fix_desc", None) is not None:
            output['fix-description'] = self.fix_desc
        if getattr(self, "fixdata", None) is not None:
            output['fix-data'] = self.fixdata

        output["verifier"] = getattr(self, "verifier", "")
        return output

    def __getstate__(self):
        """ Used by jsonpickle to state of class to output """
        return self._get_state()

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_dict(node._get_state())

# class VerifierError:

#     # This will be set as a class variable so all VerifierError instances use the same config
#     error_config = {"default-severity":"ERROR","location-text":"{}", "entry-text":"{}"}
#     values = {}
#     templated_error_texts = {}

#     def __init__(self, verifier_config:dict, verifier:str, description, location:key, errortype:ErrorType = ErrorType.NOT_SET, section:str = None, entry:str = None):

#         # TODO - Separate error text and any other output or value text into it's own file, so it can easily be changed.  Other config files when they need to use these values can reference the field name.  Sometimes the field name should be configurable e.g. in_scope_id: YES_TEXT, YES_TEXT:"Yes"
#         # TODO - include an "allowed_values" parameter, for more clearly showing what is allowed.
#         # TODO - introduce a template language (jinja) for the error texts, and pass in an object with the values (include verifier, location as well), combining them in this method.

#         #error_config = self.error_config["errors"]

#         self.verifier = verifier
#         self.description = description

#         self.errortype = errortype
#         if errortype == ErrorType.NOT_SET:
#             self.errortype = ErrorType[self.error_config["default-severity"].upper()]

#         self.section = None
#         # Find the section for the location
#         if section is not None:
#             self.section = self.error_config["location-text"].format(section)
#         if location is not None and self.section is None:
#             sectionKey = keymaster.get_section_for_key(location)
#             if sectionKey.getProperty("section") is not None:
#                 # 'section' shoud just be a string
#                 self.section = self.error_config["location-text"].format(sectionKey.getProperty("section"))

#         # Get the row-identifier for the location
#         self.entry = None
#         if entry is not None:
#             self.entry = self.error_config["entry-text"].format(entry)
#         if location is not None and entry is None:
#             row_identifier_name, row_identifier_value = keymaster.get_row_identifiers_for_key(location)
#             self.entry = self.error_config["entry-text"].format(row_identifier_name, row_identifier_value)
#             # # Find the entry in the section to help locate the error
#             # if location.hasTag("row-identifier") or location.getProperty("rowID") is not None:
#             #     # It's possible the entry key is the same as the error location, in which case the 
#             #     # "rowID" property will not be set on it (it would be a circular reference)
#             #     entryKey = location.getProperty("rowID")
#             #     if entryKey is None:
#             #         #print(f"Entry is tagged with row-identifier, property 'value' = {location.getProperty('value')}")
#             #         name = location.name
#             #         self.entry = self.error_config["entry-text"].format(name, location.getProperty("value"))
#             #     else:
#             #         #print("Entry has property rowID")
#             #         name = entryKey.name
#             #         self.entry = self.error_config["entry-text"].format(name, entryKey.getProperty("value"))
            
#     def _get_state(self):
#         output = {}
#         output['type'] = self.errortype.name
#         output['description'] = self.description
#         if self.section is not None:
#             output['section'] = self.section
#         if self.entry is not None:
#             output['entry'] = self.entry
#         output["verifier"] = self.verifier
#         return output

#     def __getstate__(self):
#         """ Used by jsonpickle to state of class to output """
#         return self._get_state()

#     def __repr__(self):
#         return pprint.pformat(self._get_state(), sort_dicts=False)



#!/usr/bin/env python3

import os
from pathlib import Path
import os.path
import logging
import convertors.gdoc_convertor.reader as reader
import convertors.html_convertor.query as query
from convertors.html_convertor.convertor import doc_to_model
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from response.response import Response

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def convert(config:dict, connection:dict, mapping:dict, doc_identifers:dict, store_doc:bool):

    # Establish connection to document location
    doc_store = reader.connect(connection)

    # TODO Check the document exists
    
    # Read the document into a string
    document = reader.read(doc_store, doc_identifers.get('id', ''))

    # Store the string
    if store_doc:
        Response.setDocument(document)

    # Read the document as html xml element
    # This will return an lxml element at the root node which is the 'html' tag
    query_document = query.get_document(document, mapping)

    # Convert the document
    return doc_to_model(config, query_document, mapping)



#!/usr/bin/env python3
"""
Loads verifiers dynamically
"""

import logging
from verifiers.verifiers_config import VerifiersConfig
from language.translate import Translate
from utils import keymaster
from utils import tags
import data.find as find
import verifiers.reference as reference

import utils.logging
from verifiers.verifier_error import VerifierIssue
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class Verifiers:
    """ Invokes a range of different configurable, extendable model verification methods

    A class that uses yaml files to configure and discover verification methods, that can then be invoked on models
    depending on the 'tags' associated with the keys of the model.
    """

    def __init__(self, config:VerifiersConfig):

        self.config = config

        VerifierIssue.issue_config = self.config.verifiers_config_dict.get("common", {}).get("errors", VerifierIssue.issue_config)
        VerifierIssue.templated_error_texts = self.config.verifiers_texts_dict.get("output-texts", {})

    def verify(self, model:dict, template_model:dict) -> list:
        """
        Executes a range of configured verifier methods on a model.

        Using the tags assigned to keys (and assigning the configured default tags to named keys), a range of verifiers are 
        invoked (can be individually disabled as well) that look for verification errors.

        Parameters
        ----------
        model : dict
            The model where the dict keys are a data.key.Key that supports having tags
        temaplate_model : dict
            The template for the model that contains known good values

        Returns
        -------
        list : A list of VerifierError objects represents the issues discovered by the invoked verifiers
        """

        self.assign_key_tags(model)
        self.assign_row_identifiers(None, model)
        self.assign_key_tags(template_model)
        self.assign_template_pre_approved(template_model)

        verifier_errors_list = []

        common_config = self.config.verifiers_config_dict["common"]
        #common_config["grouped-text"] = self.config.verifiers_texts_dict["grouped-text"]
        common_config["output-texts"] = self.config.verifiers_texts_dict["output-texts"]

        for verifier in self.config.dispatch:

            if verifier in self.config.disable:
                logger.warning(f"Verifier '{verifier}' has not been run because it was configured as disabled")

            verifier_config = self.config.verifiers_config_dict[verifier]

            logger.info(f"Entering verifier '{verifier}'")
            
            errors_list = self.config.dispatch[verifier](common_config, verifier_config, model, template_model)

            logger.info(f"Exiting verifier '{verifier}'")

            # Update errors to provide more useful output information
            for error in errors_list:
                error.verifier = verifier

            verifier_errors_list.extend(errors_list)

        return verifier_errors_list

    def assign_key_tags(self, model:dict):
        """
        Assigns default tags to data.key.Keys in a model.  Tags are the basis for verifiers.

        Using the tag mapping file configured in the verifier config (or a default), tags are (only) added
        to model keys.  It's possible to tag keys with 'no-defaults' to avoid default keys being added.

        Parameters
        ----------
        model : dict
            The model where the dict keys are a data.key.Key that supports having tags

        Returns
        -------
        Nothings.  Keys are updated in place.
        """

        tags_list = self.config.tag_mapping

        tags_dict = {}
        # Create dict of tags key-name (as key) and tags list (as value) to easily match on key-name
        for tag_entry in tags_list:
            tags_dict[Translate.localise(tag_entry["key-name"])] = tag_entry["tags"]
        
        def _assign_tags(parentKey, model):

            if isinstance(model, dict):
                for dict_key, dict_value in model.items():
                    dict_key.addProperty("parentKey", parentKey)
                    if dict_key.name in tags_dict:
                        if not dict_key.hasTag("no-defaults"):
                            # Append the tags
                            dict_key.addTags(tags_dict[dict_key.name])
                                
                    if isinstance(dict_value, dict) or isinstance(dict_value, list):
                        _assign_tags(dict_key, dict_value)

            if isinstance(model, list):
                for list_entry in model:
                    if isinstance(list_entry, dict) or isinstance(list_entry, list):
                        _assign_tags(parentKey, list_entry)
                
            return

        # Create recursive function that for each dict key checks if a tag from the file needs to be added.
        _assign_tags(None, model)

        return

    # In order to report the location of verification errors we need to tag keys with the rows they are in
    def assign_row_identifiers(self, rowIDKey, model):

        if isinstance(model, dict):
            for dict_key, dict_value in model.items():
                # We dont know where in the hierarchy we are, but for dicts we just add the rowIDKey
                if rowIDKey is not None:
                    dict_key.addProperty("rowID", rowIDKey)
                            
                if isinstance(dict_value, dict) or isinstance(dict_value, list):
                    self.assign_row_identifiers(rowIDKey, dict_value)

        if isinstance(model, list):
            # This could be a row, let's try and find a "row-identifier" tag
            for list_entry in model:
                if isinstance(list_entry, dict):
                    local_rowIDKey = rowIDKey
                    # Rows are stored as dict
                    for dict_key, dict_value in list_entry.items():
                        if dict_key.hasTag("row-identifier"):
                            #print(f"Found key tagged with 'row-identifier - {dict_key}")
                            # We found a "row-identifier" for an entry in this row, let's use that from now on
                            local_rowIDKey = dict_key
                            # We need to report errors with the value for the "row-identiifer" key, so let's store the value as a property of the key
                            dict_key.addProperty("value", dict_value)
                            # We need to be able to get sibling data in a row, so also store a reference to the whole row against the 'row-identifier' key
                            dict_key.addProperty("row", list_entry)

                            # There should only be 1 row-identifier per row
                            break
                    # Let's assign the the rowIDKey to all elements of this row (the 'row-identifier' field might not have been the first)
                    # This will include child elements as well
                    self.assign_row_identifiers(local_rowIDKey, list_entry)
                else:
                    # Not a row so recurse using existing rowIDKey
                    self.assign_row_identifiers(rowIDKey, list_entry)
            
        return


    def assign_template_pre_approved(self, template_model:dict):
        """ 
        Extracts from the template pre-approved values that can be used in the threat model
        
        A pre-approved value is a value that should be a reference to another value, but that DOESN'T verify.  The idea being that values
        that are actual references should be included in the TM, but the ones that don't should be treated as pre-approved.  Pre-approved
        values are useful so can list types of data and we don't have to list common components.
        """

        # Search the template for key ending in the template prefix
        for key_entry, value_entry in find.keys_with_tag_matching_regex(template_model, "^" + self.config.verifiers_config_dict["common"]["references"]["templ-tag-prefix"] + ".*$"):

            if value_entry == "":
                continue

            # Check whether that key value references a value in the template (in this case the template is the doc, so we use the doc prefix)
            if not reference.check_reference(template_model, self.config.verifiers_config_dict["common"]["references"]["doc-tag-prefix"], key_entry, value_entry, None, None):
                # No match.  So since this value does not correctly reference a value in the template, it must be a value added to the template
                # that the template authors want to provide as an option for TM authors to use, without having to properly reference it

                # Get the tags on this key that match the template reference
                data_section_tag = keymaster.get_data_tag_for_key(key_entry)
                matching_tags = tags.get_matching_tags(key_entry, self.config.verifiers_config_dict["common"]["references"]["templ-tag-prefix"], data_section_tag, "", "")

                # Set a property on the template key, marking it as a pre-approved value
                key_entry.addProperty("templatePreApproved", matching_tags)



#!/usr/bin/env python3
"""
Configuration class to store config for data module
"""

import logging

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


class DataConfig:
    """ Stores configuration information for converting document parts to models """

    config:dict
    _output_remove_na:bool
    _output_remove_na_tags:list

    @classmethod
    def init(cls, config:dict = {}):

        cls.config = config

        cls._output_remove_na = config.get("output", {}).get("list-with-single-not-applicable-entry", {}).get("remove", True)
        cls._output_remove_na_tags = config.get("output", {}).get("list-with-single-not-applicable-entry", {}).get("apply-tags", [])

    @classmethod
    def remove_na_single_list_entry(cls):
        """ Config for removing list entries with a single 'Not Application' value """
        return cls._output_remove_na, cls._output_remove_na_tags




#!/usr/bin/env python3
"""
Configuration loading for Convert action
"""

import logging
from pathlib import Path
from language.translate import Translate
from utils.config import ConfigBase
import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

CONVERTORS_CONFIG_YAML = "convertors_config.yaml"
CONVERTORS_CONFIG_YAML_PATH = str(Path(__file__).absolute().parent.joinpath(CONVERTORS_CONFIG_YAML))

def config() -> dict:

    yaml_config_dict = Translate.localiseYamlFile(ConfigBase.getConfigPath(CONVERTORS_CONFIG_YAML_PATH))

    return yaml_config_dict["convertors-config"]


import uvicorn
from actions.handler import lambda_handler, getVersion
from fastapi import FastAPI, Response

app = FastAPI()

def apicall(lang: str = None, format: str = None, action: str = None, scheme: str = None, docloc: str = None, meta: str = None, doctemplate: str = None, ID: str = None, IDprefix: str = None, reports: str = None):

    if format not in ['json', 'yaml', 'html']:
        format = 'json'

    if meta not in ['none', 'tags', 'properties']:
        meta = "tags"

    if reports not in ['none', 'assets', 'controls', 'all']:
        reports = "none"

    # Build input for handler
    event ={}

    class Object(object):
        pass
    context = Object()
    # setattr(context, "threatware.cli", False)
    setattr(context, "threatware.api", True)

    event["queryStringParameters"] = {}
    event["queryStringParameters"]["lang"] = lang
    event["queryStringParameters"]["format"] = format
    event["queryStringParameters"]["action"] = action
    event["queryStringParameters"]["scheme"] = scheme
    event["queryStringParameters"]["docloc"] = docloc
    event["queryStringParameters"]["meta"] = meta
    event["queryStringParameters"]["doctemplate"] = doctemplate
    event["queryStringParameters"]["ID"] = ID
    event["queryStringParameters"]["IDprefix"] = IDprefix
    event["queryStringParameters"]["reports"] = reports

    response = lambda_handler(event, context)

    return Response(content=response["body"], headers=response["headers"], status_code=response["statusCode"])

@app.get("/")
@app.get("/version/")
def version():

    return getVersion('threatware')

@app.get("/convert/")
def convert(scheme: str, docloc: str, lang: str = None, format: str = None, meta: str = None):
    action = "convert"

    return apicall(lang=lang, format=format, action=action, scheme=scheme, docloc=docloc, meta=meta)

@app.get("/verify/")
def verify(scheme: str, docloc: str, doctemplate: str, reports:str = None, lang: str = None, format: str = None, meta: str = None):
    action = "verify"

    return apicall(lang=lang, format=format, action=action, scheme=scheme, docloc=docloc, doctemplate=doctemplate, meta=meta)

@app.get("/manage/indexdata")
def manage_indexdata(ID: str, lang: str = None, format: str = None, meta: str = None):
    action = "manage.indexdata"

    return apicall(lang=lang, format=format, action=action, ID=ID, meta=meta)

@app.get("/manage/create")
def manage_create(IDprefix: str, scheme: str, docloc: str, lang: str = None, format: str = None, meta: str = None):
    action = "manage.create"

    return apicall(lang=lang, format=format, action=action, IDprefix=IDprefix, scheme=scheme, docloc=docloc, meta=meta)

@app.get("/manage/check")
def manage_check(scheme: str, docloc: str, lang: str = None, format: str = None, meta: str = None):
    action = "manage.check"

    return apicall(lang=lang, format=format, action=action, scheme=scheme, docloc=docloc, meta=meta)

@app.get("/manage/submit")
def manage_submit(scheme: str, docloc: str, lang: str = None, format: str = None, meta: str = None):
    action = "manage.submit"

    return apicall(lang=lang, format=format, action=action, scheme=scheme, docloc=docloc, meta=meta)

@app.get("/measure/")
def measure(scheme: str, docloc: str, doctemplate: str, lang: str = None, format: str = None, meta: str = None):
    action = "measure"

    return apicall(lang=lang, format=format, action=action, scheme=scheme, docloc=docloc, doctemplate=doctemplate, meta=meta)


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)


#!/usr/bin/env python3

import logging
import re
import urllib.parse
from utils.property_str import pstr

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# Remove prefix and suffix whitespace
def _trim(value):
    output = value
    
    if isinstance(value, str):
        return value.strip()

    if isinstance(value, list):
        output = [_trim(i) for i in value]

    return output

# Returns values split by different possible methods
def extract(input, query_cfg):

    if input is None:
        return None

    output = input

    if regex := query_cfg.get("regex"):
        if not isinstance(input, str):
            logger.error(f"Input was expecting string and got '{type(input)}'")
            return None
        
        if match := re.match(regex, input):
            if "group" in query_cfg:
                group = query_cfg.get("group")
            else:
                # By default, we are expecting the last group to be the matching one
                group = match.lastindex
        
            output = match.group(group)

    return pstr(_trim(output), properties = input.properties)

# Returns all values that match.  Returned values are unchanged.
def match(input, query_cfg):

    if input is None:
        return None

    output = ""

    if regex := query_cfg.get("regex"):
        if not isinstance(input, str):
            logger.error(f"Input was expecting string and got '{type(input)}'")
            return None
        if re.match(regex, input):
            # We are expecting the last group to be the matching one
            output = input

    return output

# Returns all values that match.  Returned values are unchanged.
def replace(input, query_cfg):

    if input is None:
        return None

    output = None

    if not isinstance(input, str):
        logger.error(f"Input was expecting string and got '{type(input)}'")
        return None

    output = input

    if match_value := query_cfg.get("match"):
        if replacement_value := query_cfg.get("replacement"):
            if input == match_value:
                output = pstr(replacement_value, properties = input.properties)

    return output

def url_decode(input, query_cfg):
    """ Expects input to be a string and outputs the string URL decoded """

    if input is None:
        return None

    output = None

    if not isinstance(input, str):
        logger.error(f"Input was expecting string and got '{type(input)}'")
        return None

    output = pstr(urllib.parse.unquote(input), properties = input.properties)

    return output

value_dispatch_table = {
    "value-extract":extract,
    "value-match":match,
    "value-replace":replace,
    "value-urldecode":url_decode
}


#!/usr/bin/env python3
"""
Helper methods for dealing with references
"""

import logging
import data.find as find
from data.key import key as Key
import utils.match as match
import utils.tags as tags
import utils.keymaster as keymaster

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def get_references(model, ref_type, ref_key, ref_value, callback, callback_config):
    """
    Gets all reference tag references a value that matches/compares to the passed in value

    References are returned from the location where the tag points to i.e. a data section

    Parameters
    ----------
    model : dict
        The model where the dict keys are a data.key.Key that supports having tags
    ref_type : str
        The type of reference e.g. "ref", "tref"
    ref_key : data.key.Key
        The Key with the reference tag attached to it
    ref_value : str 
        The value of the key with the reference tag
    callback : function(callback_config, comparison_str, compare_to_value, compare_value) -> bool
        A callback to do do custom comparisons between referenced values and ref_value
    callback_config : dict
        Configuration for callback

    Returns
    -------
    list(tuple) : 
        - The reference tag that references a value that matches the value.  
        - The key of the reference
        - The value of the reference
    """
    #logger.info(f"Entering get_references")
    referenced = []
    for tag in ref_key.getTags():

        tag_tuple = tags.get_quad_tag_parts(tag)
        tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

        if ref_type != tag_prefix:
            # We are only looking at tags with the desired prefix
            continue

        # Get the table to search for the value
        table_key, table_value = find.key_with_tag(model, tag_data_tag_name)

        if table_key is None:
            logger.warning(f"Reference '{tag}' included a data tag location of '{tag_data_tag_name}' which could not be found")
            continue

        result_list = find.keys_with_tag(table_value, tag_field_tag_name)

        if len(result_list) == 0:
            logger.debug(f"Reference '{tag}' included a field tag location of '{tag_field_tag_name}' which could not be found")
            continue

        for found_key, found_value in result_list:
            if tags.check_tag_comparison(tag_tuple, ref_value, found_key, found_value, callback, callback_config):
                found_referenced = (tag, found_key, found_value)
                if found_referenced not in referenced:
                    referenced.append(found_referenced)
    
    #logger.info(f"Exiting get_references")
    return referenced

def check_reference(model, ref_type, ref_key, ref_value, callback, callback_config):

    return len(get_references(model, ref_type, ref_key, ref_value, callback, callback_config)) > 0
    

def check_reference_row(row, ref_type, ref_key, ref_value, callback, callback_config, only_callback):
    """
    Checks whether a reference tag references a value in a row that matches/compares to the passed in value

    References are returned from the passed in row

    Parameters
    ----------
    row : dict
        The row data to look for the reference
    ref_type : str
        The type of reference e.g. "ref", "tref"
    ref_key : data.key.Key
        The Key with the reference tag attached to it
    ref_value : str 
        The value of the key with the reference tag
    callback : function(callback_config, tag, compare_to_value, compare_value) -> bool
        A callback to do do custom comparisons between referenced values and ref_value
    callback_config : dict
        Configuration for callback
    only_callback: bool
        Only use the callback to do the comparison, and ignore default comparison methods

    Returns
    -------
    str : The reference tag that references a value that matches the value.  'None' if no match found.
    """

    # Need a Key so we can verify the data section
    row_id_key, row_id_data = find.key_with_tag(row, "row-identifier")
    row_data_tag_name = keymaster.get_data_tag_for_key(row_id_key)

    for tag in ref_key.getTags():

        tag_tuple = tags.get_quad_tag_parts(tag)
        tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tag_tuple

        if ref_type != tag_prefix:
            # We are only looking at tags with the desired prefix
            continue

        # Check the tag_data_tag_name matches the data for the row
        if row_data_tag_name == tag_data_tag_name:

            result_list = find.keys_with_tag(row, tag_field_tag_name)

            if len(result_list) == 0:
                logger.warning(f"Reference '{tag}' included a field tag location of '{tag_field_tag_name}' which could not be found")
                continue

            for found_key, found_value in result_list:
                if tags.check_tag_comparison(tag_tuple, ref_value, found_key, found_value, callback, callback_config, only_callback):
                    return tag
    
    return None


def get_reference_descriptions(model:dict, ref_type:str, ref_key:Key):

    description = []

    # For each tag with the prefix (so if any match, it verifies)
    for tag in ref_key.getTags():

        tag_prefix, tag_data_tag_name, tag_field_tag_name, tag_comparison = tags.get_quad_tag_parts(tag)

        if ref_type != tag_prefix:
            # A key might have several other tags unrelated to references
            continue
        
        # Get the table to search for the value
        table_key, table_value = find.key_with_tag(model, tag_data_tag_name)
        if table_key is not None:
            if (sectionKey := keymaster.get_section_for_key(table_key)) is None:
                logger.warning(f"Could not find section property for '{table_key.name}'")
                section_name = table_key.name
            else:
                section_name = sectionKey.getProperty("section")
        else:
            logger.warning(f"Could not find data table with tag '{tag_data_tag_name}'")
            section_name = tag_data_tag_name
        
        field_key, _ = find.key_with_tag(table_value, tag_field_tag_name)
        field_colname = None
        if field_key is not None:
            field_colname = field_key.getProperty("colname")
        if field_colname is None:
            logger.debug(f"Could not find data field with tag '{tag_field_tag_name}'")
            field_colname = tag_field_tag_name

        if match.is_empty(tag_comparison):
            tag_comparison = "equals"

        section_field_tuple = (section_name, field_colname, tag_comparison)
        if section_field_tuple not in description:
            description.append(section_field_tuple)

    output = [{'table':tablename, 'column':colname, 'match-type':tag_comparison} for tablename, colname, tag_comparison in description]

    return output


#!/usr/bin/env python3
"""
Measure the distance between two threat models
"""


# We can tag fields with 'measure' or 'risk_analysis', and pass in a doc and a template, and compare values in these fields.  The template needn't be an actual template and could be any other TM, like a golden TM

import logging
from pickle import FALSE
import data
import measure.measure_config as manage_config
from measure.measure_output import MeasureOutput
from measure.measure_output import Measurement
from data import find
from utils import keymaster, match

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def _get_tag_from_parts(tag_tuple, exclude_filter = False):

    tag_prefix, tag_model_name, tag_compare_name, tag_filter = tag_tuple

    tag = tag_prefix + "/" + tag_model_name + "/" + tag_compare_name

    if not exclude_filter and tag_filter is not None:
        tag = tag + "/" + tag_filter

    return tag

def _get_measure_tag_parts(reftag:str, exclude_filter = False):

    # Split tag into parts
    tag_parts = reftag.split("/")

    if len(tag_parts) < 3 or len(tag_parts) > 4:
        return None, None, None, None

    tag_prefix = tag_parts[0]
    tag_model_name = tag_parts[1]
    tag_compare_name = tag_parts[2]
    tag_filter = None
    if not exclude_filter and len(tag_parts) == 4:
        tag_filter = tag_parts[3]

    return tag_prefix, tag_model_name, tag_compare_name, tag_filter

def _get_measurement_parameters(config:dict, model:dict):

    measuremeant_prefix = config["measure-tag"]["prefix"]
    measurement_tuple_keyword = config["measure-tag"]["measure-compare-tuple"]


    # Find every key with a tag that starts with the configured prefix.
    all_measure_tagged_keys = find.keys_with_tag_matching_regex(model, "^" + measuremeant_prefix + ".*$")

    targets = []

    for key_entry, value_entry in all_measure_tagged_keys:

        data_tag = keymaster.get_data_tag_for_key(key_entry)

        for tag in key_entry.getTags():

            base_tag_tuple = _get_measure_tag_parts(tag, exclude_filter=True)
            tag_prefix, tag_model_name, tag_compare_name, _ = base_tag_tuple

            if measuremeant_prefix != tag_prefix:
                # We are only looking at tags with the desired prefix
                continue

            # Each data-tag and tuple will be that same across all rows, so no need to find these more than once
            target_exists = False
            for existing_data_tag, _, existing_tag_tuple in targets:
                if existing_data_tag == data_tag and existing_tag_tuple == base_tag_tuple:
                    target_exists = True
                    break
            if target_exists:
                continue

            if measurement_tuple_keyword not in tag_compare_name:
                # Then simple add a target
                target = data_tag, [(key_entry, _get_measure_tag_parts(tag))], base_tag_tuple
                if target not in targets:
                    logger.debug(f"Adding measure target '{target}'")
                    targets.append(target)
            else:
                # Need every key name that includes 'tuple' in the row
                row_key = keymaster.get_row_identifier_for_key(key_entry)
                if row_key is None:
                    logger.warning(f"No field with tag 'row-identifier' was found in data tagged '{data_tag}'. Ignoring.")
                    continue
                row_list = row_key.getProperty("row")
                
                # Need to find matching keys that may also have a filter
                keyvalues = find.keys_with_tag_matching_regex(row_list, "^" + _get_tag_from_parts(base_tag_tuple, exclude_filter=True) + "(/.*)?$")
                key_list = []
                tag_tuple_list = []
                for dict_key, _ in keyvalues:
                    if dict_key.name not in key_list:
                        key_list.append(dict_key)
                        tag_tuple_list.append(_get_full_tag_tuple_from_base(dict_key, base_tag_tuple))

                if len(key_list) < 2:
                    logger.warning(f"A '{measuremeant_prefix}' tag with 'tuple' was used but only {len(key_list)} key entry was found in the row.  Tags with 'tuple' should identify multiple keys.")
              
                # Add the base tag_tuple, excluding any filter
                target = data_tag, list(zip(key_list, tag_tuple_list)), _get_measure_tag_parts(_get_tag_from_parts(base_tag_tuple, exclude_filter=True))
                
                if target not in targets:
                    logger.debug(f"Adding measure target '{target}'")
                    targets.append(target)

    return targets

def _get_full_tag_tuple_from_base(key, base_tag_tuple):

    for tag in key.getTags():
        key_tag_tuple = _get_measure_tag_parts(tag)
        if key_tag_tuple[0] == base_tag_tuple[0] and key_tag_tuple[1] == base_tag_tuple[1] and key_tag_tuple[2] == base_tag_tuple[2]:
            return key_tag_tuple

    logger.warning(f"Could not find base tag '{base_tag_tuple}' on key '{key}'")
    return base_tag_tuple

def _filter_row(config, key_values, base_tag_tuple):
    """ Determines if any filtered doesn't match """

    filters = config["filters"]

    for key, value in key_values:
        key_full_tag_tuple = _get_full_tag_tuple_from_base(key, base_tag_tuple)
        if (tag_filter := key_full_tag_tuple[3]) is not None:
            if tag_filter in filters:
                filter = filters[tag_filter]
                if not match.equals(value, filter):
                    return True
            else:
                logger.warning(f"Could not find filter '{tag_filter}'.  Ignoring")

    return False

def _get_row_of_measured_key_values(config, row_list, base_tag_tuple):

    tag = _get_tag_from_parts(base_tag_tuple)

    #key_values = find.keys_with_tag(row_list, tag)
    # Need to find matching keys that may also have a filter
    row_key_values = find.keys_with_tag_matching_regex(row_list, "^" + _get_tag_from_parts(base_tag_tuple, exclude_filter=True) + "(/.*)?$")

    # TODO If not a tuple, then we only want a single key/value, which should match the key passed in

    if _filter_row(config, row_key_values, base_tag_tuple):
        return None

    return row_key_values

def _get_rows_of_measured_key_values(config, data_tag_value, base_tag_tuple):

    if isinstance(data_tag_value, dict):
        data_tag_value = [data_tag_value]

    rows = []
    for row in data_tag_value:
        row_key_values = _get_row_of_measured_key_values(config, row, base_tag_tuple)
        if row_key_values is not None:
            rows.append(row_key_values)

    return rows

def distance(config:dict, measurement:Measurement, data_tag, base_tag_tuple, this_model:dict, other_model:dict):

    _, this_data_tag_value = find.key_with_tag(this_model, data_tag)
    _, other_data_tag_value = find.key_with_tag(other_model, data_tag)
    
    # Need to get all row values, then extract out the columns to compare, and then compare each column with the other.  There can be multiple columns of the same name

    this_rows = _get_rows_of_measured_key_values(config, this_data_tag_value, base_tag_tuple)
    other_rows = _get_rows_of_measured_key_values(config, other_data_tag_value, base_tag_tuple)

    measurement.addCount(len(other_rows))

    # These loops look through this_model for values that don't exist (are extra), compared to the other_model
    for this_row in this_rows:
        
        match_found = False
        for other_row in other_rows:
            # Go through every other_row, looking for match
            row_match_found = True
        
            # For every measure tag value in this_row
            for this_key, this_value in this_row:
        
                # We want to know if this_value matches a other_value
                col_match_found = False
                for other_key, other_value in other_row:
                    # The same tag could be used in multiple columns, so make sure we are comparing matching columns
                    if this_key.name == other_key.name:
                        # Check if the other_value is a match
                        # TODO account for partial matches
                        if match.equals(this_value, other_value):
                            col_match_found = True
                            break

                if not col_match_found:
                    row_match_found = False
                    break

            if row_match_found:
                match_found = True
                break
        
        if not match_found:
            # Add the 'this' value to the measurement, as it's value was not found in 'other'
            measurement.addDistance(this_row)

    return


def distances(config:dict, output:MeasureOutput, this_model:dict, other_model:dict):

    measure_from_this_model = config["measure-tag"]["measure-this-model-against-other-model"]
    measure_from_other_model = config["measure-tag"]["measure-other-model-against-this-model"]
    measure_compare_missing = config["measure-tag"]["measure-compare-missing"]
    measure_compare_extra = config["measure-tag"]["measure-compare-extra"]

    # Need to get sets of sections, columns and measure labels.  Complicated because multiple different lables can be used in a section.
    parameters = _get_measurement_parameters(config, this_model)

    this_model_name = find.key_with_tag(this_model, "document-title")[1]
    this_model_version = find.key_with_tag(this_model, "current-version")[1]
    other_model_name = find.key_with_tag(other_model, "document-title")[1]
    other_model_version = find.key_with_tag(other_model, "current-version")[1]

    for data_tag, key_tag_list, base_tag_tuple in parameters:

        # Setup the measurement capture
        data_tag_key, data_tag_value = find.key_with_tag(this_model, data_tag)
        section_key = keymaster.get_section_for_key(data_tag_key)
        measurement = Measurement(config, this_model_name, this_model_version, other_model_name, other_model_version, section_key, base_tag_tuple, key_tag_list)

        tag_prefix, tag_measure_direction, tag_compare_name, tag_filter = base_tag_tuple
        # Calculate the distance
        #if match.equals(tag_measure_direction, measure_from_this_model):
        if match.equals(tag_measure_direction, measure_from_other_model) and measure_compare_missing in tag_compare_name or \
            match.equals(tag_measure_direction, measure_from_this_model) and measure_compare_extra in tag_compare_name:
            # measure/from-other/missing is equivalent to measure/from-this/extra
            distance(config, measurement, data_tag, base_tag_tuple, this_model, other_model)
        #elif match.equals(tag_measure_direction, measure_from_other_model):
        elif match.equals(tag_measure_direction, measure_from_this_model) and measure_compare_missing in tag_compare_name or \
            match.equals(tag_measure_direction, measure_from_other_model) and measure_compare_extra in tag_compare_name:
            # measure/from-this/missing is equivalent to measure/from-other/extra
            # Note, we simple switch the order of the models we pass in
            distance(config, measurement, data_tag, base_tag_tuple, other_model, this_model)
        else:
            # TODO capture error in measurement output
            logger.warning(f"Measure tag found referencing unknown model '{tag_measure_direction}'.  Ignoring")
            continue

        output.addMeasure(measurement)

    output.setSuccess("success", {"this_model_name":this_model_name, "other_model_name":other_model_name})
    return output    
    




#!/usr/bin/env python3
"""
Validates whether a value matches a value from the template
"""

import logging
from pathlib import Path
import data.find as find
import utils.match as match
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def validate(config:dict, key:Key, value:str, references:dict) -> bool:
    
    # Need to get some config from references
    if not (key_tag := references['validator-tag']):
        logger.error("The references parameter did not have a 'validator-tag' key")
        return False

    if not (template_model := references['template-model']):
        logger.error("The references parameter did not have a 'template-model' key")
        return False
    
    tagged_data = find.keys_with_tag(template_model, key_tag)
    
    match_found = False
    for tagged_key, tagged_value in tagged_data:

        # Lots of keys could be tagged as being validated from template, but we only want to look at template values for keys with the same name as the one
        # we are validating (could still possibly lead to issues)
        if tagged_key == key and match.equals(value, tagged_value):
            match_found = True
            break
    
    #if not match_found:
    #    tagged_values = [tagged_value for _, tagged_value in tagged_data]
    #    key.addProperty(references['validator-tag'], config["output_text_invalid"].format(key, value, tagged_values))

    return match_found


#!/usr/bin/env python3

import logging
from atlassian import Confluence
from requests import HTTPError
from utils.error import ConvertError
import unicodedata

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def connect(connection:dict):

    confluence = None

    if (connection.get("cloud", "True")).lower() == "true":

        # For connecting to Cloud Confluence
        confluence = Confluence(
            url=connection["url"],
            username=connection["username"],
            password=connection["api_token"],
            cloud=True)

    elif "password" in connection:

        # For connecting to Server or Cloud Confluence (but don't use passwords! for this)
        confluence = Confluence(
            url=connection["url"],
            username=connection["username"],
            password=connection["password"])

    elif "token" in connection:

        # For connecting to Server Confluence using Personal Access Token (PAT)
        confluence = Confluence(
            url=connection["url"],
            token=connection["token"])
    
    return confluence

def exists(confluence, space, title):
    return confluence.page_exists(space, title)

def exists(confluence, page_id):
    
    url = "rest/api/content/{}/property".format(page_id)
    
    try:
        confluence.get(url)
    except HTTPError as e:
        logger.error(e)
        if e.response.status_code == 404:
            raise ConvertError("not-found", {"ID":page_id, "url":e.response.url})
        raise

    return True

def history(confluence, page_id):
    url = "rest/api/content/{}/version?expand=content".format(page_id)
    return confluence.get(url)

def read(confluence, page_id, version=''):
    
    page = confluence.get_page_by_id(page_id, expand='body.view', status=None, version=version)

    return unicodedata.normalize("NFKD", page['body']['view']['value'])


#!/usr/bin/env python3

import logging
from utils.load_yaml import yaml_register_class
from enum import Enum

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class KeySerialiseType(Enum):
    NO_TAGS_PROPERTIES = 0
    TAGS = 1
    TAGS_PROPERTIES = 2


class key:
    yaml_tag = u'!Key'
    
    serialise_type:KeySerialiseType = KeySerialiseType.TAGS_PROPERTIES

    def __init__(self, name:str, tags:list = None):
        # Fun python fact, don't have default value be something you want to change e.g. list, 
        # because default values are initialised once so all class instances will share them
        self.name = name
        if tags is None:
            self.tags = list()
        else:   
            self.tags = list(tags)  # This is apparently key otherwise we will edit the passed in list 
        self.properties = {}

        yaml_register_class(key)

    def copy(obj):
        if isinstance(obj, key):
            return key(obj.name, obj.tags)
        return None

    def addTag(self, tag:str):
        if not tag in self.tags:
            self.tags.append(tag)

    def addTags(self, tags:list):
        for tag in tags:
            self.addTag(tag)

    def hasTag(self, tag):
        return tag in self.tags

    def getTags(self):
        return self.tags

    def addProperty(self, property_name, property_value):
        if property_name is not None and isinstance(property_name, str) and property_name != "":
            self.properties[property_name] = property_value

    def getProperty(self, property_name):
        return self.properties.get(property_name, None)

    def __str__(self):
        return self.name
    
    # def __getstate__(self):
    #     # To customise jsonpickle output
    #     return {"name":self.__str__()}

    # jsonpickle has a parameter 'keys' (by default False) that when it comes across non-string dictionary keys, like how we use this class, it will call
    # repr() on the object to get a string version.
    def __repr__(self):
        if key.serialise_type == KeySerialiseType.NO_TAGS_PROPERTIES:
            return self.name
        #elif key.serialise_type == KeySerialiseType.TAGS or key.serialise_type == KeySerialiseType.TAGS_PROPERTIES:
        return f"<name:{self.name}, tags:{self.tags}>"
        # if not self.tags:
        #     return self.name
        # else:
        #     return f"<name:{self.name}, tags:{self.tags}>"

    def __eq__(self, other):

        if isinstance(other, str):
            return self.name == other

        if isinstance(other, key):
            return (self.name == other.name)
            #return (self.name == other.name) and (self.tags.sort() == other.tags.sort())

        return NotImplemented

    def __hash__(self):
        return hash(self.name)

    @classmethod
    def config_serialisation(cls, meta_level:str):
         
        if meta_level == "none":
            cls.serialise_type = KeySerialiseType.NO_TAGS_PROPERTIES
        elif meta_level == "tags":
            cls.serialise_type = KeySerialiseType.TAGS
        elif meta_level == "properties":
            cls.serialise_type = KeySerialiseType.TAGS_PROPERTIES

    @classmethod
    def to_yaml(cls, representer, node):
        if cls.serialise_type == KeySerialiseType.NO_TAGS_PROPERTIES:
            return representer.represent_str(node.name)
        elif cls.serialise_type == KeySerialiseType.TAGS:
            return representer.represent_mapping(cls.yaml_tag, {"name":node.name, "tags":node.tags})
        elif cls.serialise_type == KeySerialiseType.TAGS_PROPERTIES:
            return representer.represent_mapping(cls.yaml_tag, {"name":node.name, "tags":node.tags, "properties":node.properties})

    @classmethod
    def from_yaml(cls, constructor, node):
        if cls.serialise_type == KeySerialiseType.NO_TAGS_PROPERTIES:
            
            logger.error(f"Deserialisation from a serialised Key that did not not include tags or properties should not be called.")
            return cls(str(constructor.construct_yaml_str(node)))

        elif cls.serialise_type == KeySerialiseType.TAGS:
            
            logger.error(f"Deserialisation from a serialised Key that did not not include properties should not be called.")

            key_dict = constructor.construct_mapping(node)

            key_obj = cls(key_dict["name"], key_dict["tags"])

            return key_obj

        elif cls.serialise_type == KeySerialiseType.TAGS_PROPERTIES:
            
            key_dict = constructor.construct_mapping(node)

            key_obj = cls(key_dict["name"], key_dict["tags"])

            for prop_name, prop_value in key_dict["properties"].items():
                key_obj.addProperty(prop_name, prop_value)

            return key_obj


#!/usr/bin/env python3
"""
Validates whether a value exists within a given set of tagged data
"""

import logging
import data.find as find
import utils.match as match
from data.key import key as Key

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def validate(config:dict, key:Key, value:str, references:dict) -> bool:
    
    # By default look in the threat model
    if not (model := references['model']):
        logger.error("The references parameter did not have a 'model' key")
        return False
    
    if config.get("template", False):
        # If 'template: True' then look in the template
        if not (model := references['template-model']):
            logger.error("The references parameter did not have a 'template-model' key")
            return False

    # If a 'data-tag' has been set then only look at that data
    if (data_tag := config.get("data-tag", None)) is not None:
        _, model = find.key_with_tag(model, data_tag)

    # Get the key to look for in the model
    if (key_tag := config.get("value-tag", None)) is None:
        logger.error("The config did not have a 'value-tag' key")
        return False
    
    tagged_data = find.keys_with_tag(model, key_tag)
    
    if match.equals(value, [key_value for key_name, key_value in tagged_data]):
        return True
    
    return False


#!/usr/bin/env python3
"""
Verifies that all keys tagged as mandatory have a non-empty value
"""

from data import find
import logging
from verifiers.verifier_error import ErrorType
from verifiers.verifier_error import VerifierIssue
from utils import match
from utils.model import recurse

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def verify(common_config:dict, verifier_config:dict, model:dict, template_model:dict) -> list:
    
    verify_return_list = []

    if verifier_config.get("default", True):

        def _inner_is_empty(key, value, context):
            is_empty = False
            if not key.hasTag(verifier_config["exceptions-tag"]):
                if value is None:
                    is_empty = True
                if isinstance(value, str):
                    if match.is_empty(value):
                        is_empty = True
                if isinstance(value, dict) or isinstance(value, list):
                    if len(value) == 0:
                        is_empty = True
            
            if is_empty:
                issue_dict = {}
                issue_dict["issue_key"] = key
                issue_dict["issue_value"] = value
                verify_return_list.append(VerifierIssue(
                    error_text_key="missing-mandatory", 
                    fix_text_key=None, 
                    issue_dict=issue_dict))
                return True, True
            else:
                return True, False

        recurse(model, _inner_is_empty, None)

    else:
        # Need to just those keys marked as "mandatory"

        tagged_data = find.keys_with_tag(model, "mandatory")

        for tagged_key, tagged_value in tagged_data:

            if match.is_empty(tagged_value):
                issue_dict = {}
                issue_dict["issue_key"] = tagged_key
                issue_dict["issue_value"] = tagged_value
                verify_return_list.append(VerifierIssue(
                    error_text_key="missing-mandatory", 
                    fix_text_key=None, 
                    issue_dict=issue_dict))


    return verify_return_list



#!/usr/bin/env python3
"""
Measure a threat model
"""


# We can tag fields with 'measure' or 'risk_analysis', and pass in a doc and a template, and compare values in these fields.  The template needn't be an actual template and could be any other TM, like a golden TM

import logging
import measure.measure_config as measure_config
import measure.measure_distance as measure_distance
from measure.measure_output import MeasureOutput
from utils.model import assign_row_identifiers, assign_parent_keys
import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))


def config():

    return measure_config.config()

def _output(config:dict):

    return MeasureOutput(config.get("output", {}))

def distance(config:dict, doc_model:dict, ref_model:dict):

    output = _output(config)

    assign_row_identifiers(doc_model)
    assign_parent_keys(doc_model)
    assign_row_identifiers(ref_model)
    assign_parent_keys(ref_model)

    return measure_distance.distances(config, output, doc_model, ref_model)

def distance_to_approved(config:dict, output:MeasureOutput, doc_model:dict):
    """ 
    Measure the distance of the passed in threat model to the approved version

    """

    # Get the ID of the TM from the model

    # Get the approved version from the index metadata

    # Get the approved version and hydrate it (not sure how to do this?)

    # Measure distance between two

def distance_to_template(config:dict, execution_env, doc_model:dict, template:dict):
    """ 
    Measure the distance of the passed in threat model to a template

    """

    output = _output(config)

    assign_row_identifiers(doc_model)
    assign_parent_keys(doc_model)
    assign_row_identifiers(template)
    assign_parent_keys(template)

    return measure_distance.distances(config, output, doc_model, template)


#!/usr/bin/env python3

import os, io
import logging
from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload
from googleapiclient.errors import HttpError
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.auth.exceptions import RefreshError
from google.oauth2.credentials import Credentials
from google.oauth2.service_account import Credentials as SvcAcctCredentials
from utils.error import ConvertError
from utils.config import ConfigBase
import httplib2

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# If modifying these scopes, delete the file token.json.
#SCOPES = ['https://www.googleapis.com/auth/documents.readonly', 'https://www.googleapis.com/auth/drive.readonly']
SCOPES = ['https://www.googleapis.com/auth/drive.readonly']
# I would like to use the below scope, but it doesn't seem possible with a CLI app
#SCOPES = ['https://www.googleapis.com/auth/drive.readonly']

def _getCredentials(connection):

    creds_dict ={}
    if "credentials-file" in connection:
        # Use the credentials file to create credentials
    
        creds = None
        token_path = ConfigBase.getConfigPath(connection.get("token-file"))
        credentials_path = ConfigBase.getConfigPath(connection.get("credentials-file"))

        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(token_path):
            creds = Credentials.from_authorized_user_file(token_path, SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            runOAuthAgain = False
            if creds and creds.expired and creds.refresh_token:
                try:
                    creds.refresh(Request())
                except(RefreshError):
                    runOAuthAgain = True
            if not creds or runOAuthAgain:
                if os.path.exists(credentials_path):
                    flow = InstalledAppFlow.from_client_secrets_file(credentials_path, SCOPES)
                    creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            if creds:
                with open(token_path, 'w') as token:
                    token.write(creds.to_json())

        creds_dict = {"credentials": creds}

    elif "api_key" in connection:
        creds_dict = {"developerKey":connection.get("api_key")}
    else:

        creds_dict = {"credentials":SvcAcctCredentials.from_service_account_info(connection, scopes=SCOPES)}

    return creds_dict

def connect(connection:dict):

    creds_dict = _getCredentials(connection)

    if not creds_dict:
        logger.error("Could not retrieve Google Doc credentials")
        raise ConvertError("internal-error", {})
    
    service = build('drive', 'v3', **creds_dict)

    return service

def exists(cservice, doc_id):
    
    return True

def history(confluence, page_id):
    
    return None

def read(service, doc_id, version=''):
    
    try: 
        request = service.files().export_media(fileId=doc_id, mimeType='text/html')

        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logger.debug("Download %d%%." % int(status.progress() * 100))
        
        document = fh.getvalue().decode('utf-8')

    except HttpError as e:
        logger.error(e)
        if e.status_code == 404:
            raise ConvertError("not-found", {"ID":doc_id, "url":e.uri})
        raise

    return document




#!/usr/bin/env python3

import logging
import convertors.html_convertor.query
import utils.text_query
import utils.value_query

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def _process_query(query_def, input):

    query_type = query_def.get("type")

    if query_type in convertors.html_convertor.query.html_dispatch_table:
        # Get document content
        content = convertors.html_convertor.query.html_dispatch_table[query_type](input, query_def)
        return content

    if query_type in utils.text_query.text_dispatch_table:
        # Get document content
        content = utils.text_query.text_dispatch_table[query_type](input, query_def)
        return content
    
    if query_type in utils.value_query.value_dispatch_table:
        # Get document content
        content = utils.value_query.value_dispatch_table[query_type](input, query_def)
        return content

    logger.error(f"Could not process query of type '{query_type}'")
    return None

# The output of processing content is always a list or value
def parse(get_data_def, input):

    logger.debug(f"Entering get.parse to process {get_data_def}")

    get_data_defs = []

    if isinstance(get_data_def, dict) and get_data_def:
        # We have just 1 query
        #if (query := get_data_def.get("query")):    # If None is returned then it evaluates to False
        get_data_defs.append(get_data_def)
    elif isinstance(get_data_def, list) and get_data_def:
        get_data_defs = get_data_def

    # Process each query.  Note, if we have no queries, we return the input
    logger.debug(f"Executing {len(get_data_defs)} queries")
    data = input
    for query_def in get_data_defs:
        data = _process_query(query_def.get("query", {}), data)
        # data_value = data
        # if isinstance(data, dict) and "value" in data:
        #     data_value = data["value"]
        
        # new_data = _process_query(query_def.get("query", {}), data_value)

        # if isinstance(data, dict) and "value" in data:
        #     data["value"] = new_data
        # else:
        #     data = new_data


    logger.debug(f"Leaving get.parse having processed {get_data_def}")

    return data


#!/usr/bin/env python3

import logging
import re
from utils.property_str import pstr
from utils.match import is_empty
import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

# Remove prefix and suffix whitespace
def _trim(value):
    output = value
    
    if isinstance(value, str):
        return value.strip()

    if isinstance(value, list):
        output = [_trim(i) for i in value]

    return output

# Returns values split by different possible methods
def split(text_value, query_cfg):

    output = text_value

    # If the output has values, but some are empty, we want to ignore those, this is done by the len(i) > 0 condition in the return
    # If the input is empty, we want to return a list with a single empty string so we have properties (of the value) to pass to the key e.g. location
    if is_empty(output):
        return [pstr("", properties = text_value.properties)]

    if splitby := query_cfg.get("split-by"):
    
        if char_separators := splitby.get("char-separators", []):
            regexPattern = '|'.join(char_separators)
            output = re.split(regexPattern, text_value)

    if lineregex := query_cfg.get("line-regex"):
        if not isinstance(text_value, str):
            logger.error(f"input was expecting string and got '{type(text_value)}'")
            return None
        output = []
        for line in text_value.splitlines():
            if match := re.match(lineregex, line):
                if "group" in query_cfg:
                    group = query_cfg.get("group")
                else:
                    # By default, we are expecting the last group to be the matching one
                    group = match.lastindex
            
                output.append(match.group(group))

    #return [i for i in _trim(output) if len(i) > 0]
    
    # Maintain the properties of the input pstr in the output
    return [pstr(i, properties = text_value.properties) for i in _trim(output) if len(i) > 0]

# Returns all values that match.  Returned values are unchanged.
def match(text_value, query_cfg):

    output = ""

    if lineregex := query_cfg.get("line-regex", {}):
        if not isinstance(text_value, str):
            logger.error(f"input was expecting string and got '{type(text_value)}'")
            return None
        output = []
        for line in text_value.splitlines():
            if re.match(lineregex, line):
                # We are expecting the last group to be the matching one
                output.append(line)

    return output

# Returns all values that match.  Returned values are unchanged.
def replace(input, query_cfg):

    output = None

    if isinstance(input, list):
        output = [replace(i, query_cfg) for i in input]

    if isinstance(input, str):
        output = input

        if match_value := query_cfg.get("match"):
            if replacement_value := query_cfg.get("replacement"):
                if input == match_value:
                    output = replacement_value

    return output

text_dispatch_table = {
    "text-split": split,
    "text-match": match,
    "text-replace": replace
}